import abc
from collections import defaultdict
import collections
import typing
from collections import deque
import dataclasses
import inspect
import json
from dataclasses import is_dataclass
from typing import get_origin, get_args, Any, Mapping, MutableMapping, OrderedDict, Literal, Union, get_type_hints, \
    Type, Sequence, Tuple, Optional

from pydantic import BaseModel, create_model
import datetime

class Validator:

    def __init__(self):
        # Extract types from collections and collections.abc
        collection_types = {cls for name, cls in collections.__dict__.items() if isinstance(cls, type)}
        abc_collection_types = {cls for name, cls in collections.abc.__dict__.items() if isinstance(cls, type)}
        # Filter out types that have dictionary-like methods
        self.dict_like_types = {
            cls for cls in collection_types.union(abc_collection_types)
            if hasattr(cls, 'keys') and hasattr(cls, 'items')
        }

        self.list_like_types = {
            cls for cls in collection_types.union(abc_collection_types)
            if hasattr(cls, 'append') and hasattr(cls, 'pop')
        }

        self.set_like_types = {
            cls for cls in collection_types.union(abc_collection_types)
            if hasattr(cls, 'add') and hasattr(cls, 'discard')
        }
        # Add the general Sequence to list-like types
        # if python version is 3.9 or above, use collections.abc.Sequence
        if hasattr(collections.abc, 'Sequence'):
            self.list_like_types.add(collections.abc.Sequence)
        else:
            self.list_like_types.add(collections.Sequence)

        self.list_like_types.add(typing.List)

        # Add the general Mapping to dict-like types
        if hasattr(collections.abc, 'Mapping'):
            self.dict_like_types.add(collections.abc.Mapping)
        else:
            self.dict_like_types.add(collections.Mapping)

        self.dict_like_types.add(typing.Dict)

        # Add the general Set to set-like types
        if hasattr(collections.abc, 'Set'):
            self.set_like_types.add(collections.abc.Set)
        else:
            self.set_like_types.add(collections.Set)

        self.set_like_types.add(typing.Set)

        # Add the general Tuple to tuple-like types
        self.tuple_like_types = {
            cls for cls in collection_types.union(abc_collection_types)
            if hasattr(cls, '__getitem__') and hasattr(cls, '__len__')
        }

        self.tuple_like_types.add(typing.Tuple)

    def is_base_type(self, _type: Any) -> bool:
        """Determine if a type is a base type."""
        return _type in {int, float, str, bool, None}

    def validate_base_type(self, value: Any, typ: Any) -> bool:
        """Validate base types."""
        if typ is None:
            return value is None
        return isinstance(value, typ)

    def validate_output(self, output: str, type_definition: Any) -> bool:
        try:
            deserialized_output = json.loads(output)
        except json.JSONDecodeError:
            return False

        return self.check_type(deserialized_output, type_definition)

    def check_type(self, value: Any, type_definition: Any) -> bool:
        """
        Validate a value against a type definition.

        Args:
            value: Any object or primitive value
            type_definition: The type definition to validate against

        Returns:
            Whether the value is valid for the type definition
        """
        if type_definition is Any:
            return True

        if self.is_base_type(type_definition):
            return self.validate_base_type(value, type_definition)

        origin = get_origin(type_definition) or type_definition
        args = get_args(type_definition)

        # Handle base types
        if self.is_base_type(origin):
            return self.validate_base_type(value, origin)

        if origin == Literal:
            return value in args

        if origin == Union:
            return any(self.check_type(value, union_type) for union_type in args)

        # Handle tuples
        if origin == tuple:
            if not isinstance(value, tuple) or (args and len(value) != len(args)):
                return False
            return all(self.check_type(v, t) for v, t in zip(value, args))

        # Handle lists
        if origin == list:
            if not isinstance(value, list):
                return False
            item_type = args[0] if args else Any
            return all(self.check_type(v, item_type) for v in value)

        # Handle more complex types that are collections and list-like
        if origin is list or issubclass(origin, tuple(self.list_like_types)):
            if not any(isinstance(value, t) for t in self.list_like_types):
                return False
            item_type = args[0] if args else Any
            return all(self.check_type(v, item_type) for v in value)

        # Handle sets
        if origin == set:
            if not isinstance(value, set):
                return False
            item_type = args[0] if args else Any
            return all(self.check_type(v, item_type) for v in value)
        
        # Handle datetime
        if origin == datetime.datetime:
            # try to instantiate datetime
            try:
                obj = datetime.datetime(**value)
                return True
            except:
                return False
        # Handle dictionaries
        if origin is dict or issubclass(origin, tuple(self.dict_like_types)):
            if not isinstance(value, (dict, Mapping)):#, MutableMapping, OrderedDict)):
                return False

            if args:
                if len(args) == 1:
                    key_type = args[0]
                    value_type = Any  # General assumption; specific dict-like types might differ
                elif len(args) == 2:
                    key_type, value_type = args
                else:
                    key_type = value_type = Any
            else:
                key_type = value_type = Any

            return all(
                self.check_type(k, key_type) and self.check_type(v, value_type)
                for k, v in value.items()
            )

        # Handle pydantic models
        if self.is_pydantic_model(origin):
            try:
                #temp_model = create_model('TempModel', **value)
                #return isinstance(temp_model, origin)
                # check if value is dict
                if not isinstance(value, dict):
                    return False
                # get all required init arguments for origin
                # required arguments are the ones withouyt default values
                required_fields = [field for field, field_type in origin.__annotations__.items() if not (typing.get_origin(field_type) is Union and type(None) in typing.get_args(field_type))]
                # check that all required arguments are in value and do type checking
                for arg in required_fields:
                    # check if it is in value
                    if arg not in value:
                        return False
                    # get the type of the argument
                    arg_type = origin.__annotations__[arg]
                    if not self.check_type(value[arg], arg_type):
                        return False
                # check that all arguments in value are correct type
                # this is additional check, because the above check only checks required arguments
                for arg, obj in value.items():
                    if arg in required_fields:
                        continue
                    arg_type = origin.__annotations__[arg]
                    if not self.check_type(value[arg], arg_type):
                        return False
                
                #origin.parse_obj(value)
                return True
            except Exception as e:
                print(e)
                return False

        # Handle dataclasses
        if self.is_dataclass_instance(origin):
            try:
                # for field in dataclasses.fields(origin):
                #     field_name = field.name
                #     field_type = field.type
                #     if field_name not in value or not self.check_type(value[field_name], field_type):
                #         return False
                # return True
                obj = origin(**value)
                return dataclasses.asdict(obj) == value
            except:
                return False

        # Handle dataclasses and arbitrary class types
        if inspect.isclass(origin) and not self.is_base_type(origin):
            # Ensure the value is an instance of the class
            if not isinstance(value, origin):
                return False

            # Gather type hints from the class and its bases
            type_hints = {}
            for cls in reversed(origin.__mro__):
                type_hints.update(get_type_hints(cls))

            # Validate each attribute of the class
            for attr, attr_type in type_hints.items():
                attr_value = getattr(value, attr, None)
                if not self.check_type(attr_value, attr_type):
                    return False

            return True

        return False

    @staticmethod
    def is_pydantic_model(cls):
        return hasattr(cls, 'parse_obj')

    @staticmethod
    def is_dataclass_instance(cls):
        return hasattr(cls, '__annotations__') and hasattr(cls, '__dataclass_fields__')


    @staticmethod
    def _is_subclass_of_generic(cls: Type, generic: Type) -> bool:
        """Determine if the class is a subclass of a generic type."""
        try:
            return issubclass(cls, generic) and cls is not generic
        except TypeError:
            if not hasattr(cls, '__origin__'):
                return False
        return cls.__origin__ is generic

    @staticmethod
    def _is_generic(cls: Type) -> bool:
        """Check if the provided type is a generic."""
        return hasattr(cls, "__origin__")

    def _get_recursive_args(self, target_type: Type) -> Tuple[Type, ...]:
        """
        Recursively check the base classes (i.e., the superclass chain) of the target type until we find one that
        retains the type arguments.
        :return: Type chain
        """
        if get_args(target_type):
            return get_args(target_type)
        for base in target_type.__bases__:
            args = self._get_recursive_args(base)
            if args:
                return args
        return ()

    def _find_generic_base_and_args(self, target_type: Type) -> Tuple[Type, Tuple[Type, ...]]:
        """
        Navigate up the MRO to find the first generic base and its arguments.
        """
        # First, check if target_type is a type annotation.
        # If so, directly return its origin and arguments.
        origin = get_origin(target_type)
        args = get_args(target_type)
        if origin and args:
            return origin, args

        # If target_type is a real class, then navigate its MRO.
        if hasattr(target_type, '__mro__'):
            if hasattr(target_type, '__orig_bases__'):
                for base in target_type.__orig_bases__:
                    if get_args(base):
                        return base, get_args(base)

            for base in target_type.__mro__:
                if get_args(base):
                    return base, get_args(base)

        return None, ()

    def _is_list_like(self, target_type: Type) -> bool:
        """Determine if the target type is list-like."""
        if target_type in {list, typing.List}:
            return True
        if hasattr(target_type, "__origin__") and target_type.__origin__ in {list, typing.List}:
            return True
        return False

    def _is_tuple_like(self, target_type: Type) -> bool:
        """Determine if the target type is tuple-like."""
        if target_type in {tuple, typing.Tuple}:
            return True
        if hasattr(target_type, "__origin__") and target_type.__origin__ in {tuple, typing.Tuple}:
            return True
        return False

    def _is_dict_like(self, target_type: Type) -> bool:
        """Determine if the target type is dict-like."""
        if target_type in {dict, typing.Dict}:
            return True
        if hasattr(target_type, "__origin__") and target_type.__origin__ in {dict, typing.Dict}:
            return True
        return False

    def _is_set_like(self, target_type: Type) -> bool:
        """Determine if the target type is set-like."""
        if target_type in {set, typing.Set}:
            return True
        if hasattr(target_type, "__origin__") and target_type.__origin__ in {set, typing.Set}:
            return True
        return False

    def instantiate(self, data: Any, target_type: Type) -> Any:
        """
        Attempts to convert a JSON-compatible data structure into an instance of the specified type.

        Args:
            data: JSON-compatible data structure to instantiate the target type.
            target_type: The type to instantiate from the given data.

        Returns:
            An instance of the target type initialized with the data.
        """

        # Handle None type
        if data is None:
            return None

        origin = get_origin(target_type) or target_type

        # If the target type is a built-in, attempt to instantiate and return
        if self.is_base_type(target_type) or target_type is Any:
            # If the parsed data is a string and target type is str, return it directly
            if isinstance(data, str) and target_type is str:
                return data

            # If any, return the data directly
            if target_type is Any:
                return data

            try:
                return target_type(data)
            except (ValueError, TypeError):
                # Handle the special case where the string represents a float but we want an integer
                if target_type is int:
                    try:
                        return int(float(data))
                    except (ValueError, TypeError):
                        pass

                if target_type is float:
                    try:
                        return int(float(data))
                    except (ValueError, TypeError):
                        pass
                raise TypeError(f"Failed to instantiate {target_type} from provided data.")
        # special handling for datetime
        if origin == datetime.datetime:
            # try to instantiate datetime
            try:
                return datetime.datetime(**data)
            except:
                raise TypeError(f"Failed to instantiate {target_type} from provided data.")

        # check if origin is Union, if so, instantiate the first type that works
        if origin == Union:
            for arg in get_args(target_type):
                try:
                    return self.instantiate(data, arg)
                except:
                    continue
            raise TypeError(f"Failed to instantiate {target_type} from provided data.")

        # If the data is a dictionary and the target is a custom class that can be instantiated from a dictionary.
        if isinstance(data, dict):
            if inspect.isclass(target_type) and not self.is_base_type(target_type):
                # Special handling for dataclasses
                if is_dataclass(target_type):
                    fields = [f.name for f in dataclasses.fields(target_type)]
                    type_hints = get_type_hints(target_type)
                    filtered_data = {k: self.instantiate(v, type_hints.get(k, Any)) for k, v in data.items() if
                                     k in fields}
                    return target_type(**filtered_data)

                # Special handling for Pydantic models
                if issubclass(target_type, BaseModel):
                    # instantiate the sub attributes
                    for attr, attr_type in target_type.__annotations__.items():
                        if attr in data:
                            data[attr] = self.instantiate(data[attr], attr_type)
                    return target_type.model_validate(data)

                # For general classes, attempt instantiation
                try:
                    return target_type(**data)
                except TypeError:
                    raise TypeError(f"Failed to instantiate {target_type.__name__} from dictionary.")
            # Handle dictionary-like types
            # Check if the target type is or inherits from defaultdict
            if origin is defaultdict or (isinstance(origin, type) and issubclass(origin, defaultdict)):
                key_type, value_type = get_args(target_type) if get_args(target_type) else (Any, Any)
                instantiated_items = {self.instantiate(k, key_type): self.instantiate(v, value_type) for k, v in
                                      data.items()}

                # For defaultdict, you'll need a default factory. Here, I'm using `int` for simplicity,
                # but you might want to adapt this based on your needs.
                return defaultdict(int, instantiated_items)

            # Handle set-like dict types like OrderedDict
            # the first check needs to be done to ensure origin has the __mro__ attribute
            elif inspect.isclass(origin)and any(issubclass(base, dict) for base in origin.__mro__):
                key_type, value_type = get_args(target_type) if get_args(target_type) else (Any, Any)
                instantiated_items = {self.instantiate(k, key_type): self.instantiate(v, value_type) for k, v in data.items()}
                return origin(instantiated_items)
            
            # Handle other dictionary-like types
            elif origin is dict or self._is_subclass_of_generic(origin, dict):
                key_type, value_type = get_args(target_type) if get_args(target_type) else (Any, Any)
                instantiated_dict = {self.instantiate(k, key_type): self.instantiate(v, value_type) for k, v in
                                     data.items()}

                # If the target_type is a subclass of dict, return an instance of target_type
                if self._is_subclass_of_generic(target_type, dict) and not self._is_generic(target_type):
                    return target_type(instantiated_dict)
                else:
                    return dict(instantiated_dict)

        # Tuples aren't supported in JSONable types, so we look for lists instead
        if isinstance(data, list):
            try:
                # If the origin or target type is a list-like type, or if it implements a list-like collections type
                # e.g Sequence[int]
                if origin is list or self._is_subclass_of_generic(origin, list):
                    base, item_types = self._find_generic_base_and_args(target_type)

                    item_type = item_types[0] if item_types else Any

                    instantiated_items = []

                    for item in data:
                        # For each item, validate and instantiate it
                        try:
                            instantiated_item = self.instantiate(item, item_type)
                        except ValueError:
                            raise TypeError(
                                f"Item of type {type(item).__name__} does not match expected type {item_type[0].__name__}.")

                        safe = self.check_type(instantiated_item, item_type)
                        if not safe:
                            raise TypeError(
                                f"Item of type {type(item).__name__} does not match expected type {item_type[0].__name__}.")
                        instantiated_items.append(instantiated_item)

                    # If target_type is a subclass of list, return an instance of target_type
                    if self._is_subclass_of_generic(target_type, list) and not self._is_generic(target_type):
                        return target_type(instantiated_items)

                # Handle tuples
                if self._is_tuple_like(target_type) or (isinstance(origin, type) and issubclass(origin, tuple)):
                    base, item_types = self._find_generic_base_and_args(target_type)

                    instantiated_items = []

                    # If there are no subscripted types, assume Any
                    if not item_types:
                        item_types = (Any,) * len(data)

                    for i, item in enumerate(data):
                        # For each item, validate and instantiate it
                        instantiated_item = self.instantiate(item, item_types[i])
                        instantiated_items.append(instantiated_item)

                        # If the instantiated item does not match the expected type, raise an exception
                        _type = item_types[i]
                        if not isinstance(instantiated_item, _type):
                            raise TypeError(
                                f"Item {i} of type {type(item).__name__} does not match expected type {item_types[i].__name__}.")

                    # Convert the list of instantiated items to a tuple
                    instantiated_tuple = tuple(instantiated_items)
                    # If target_type is a subclass of tuple, return an instance of target_type
                    if self._is_subclass_of_generic(target_type, tuple):
                        return target_type(instantiated_tuple)

                    return instantiated_tuple

                # Handle sets
                if self._is_set_like(target_type) or (isinstance(origin, type) and issubclass(origin, set)):
                    base, item_type = self._find_generic_base_and_args(target_type)

                    if not item_type:
                        item_type = Any

                    instantiated_items = set()

                    for item in data:
                        # For each item, validate and instantiate it
                        instantiated_item = self.instantiate(item, item_type[0])
                        instantiated_items.add(instantiated_item)

                        # If the instantiated item does not match the expected type, raise an exception
                        if not isinstance(instantiated_item, item_type[0]):
                            raise TypeError(
                                f"Item of type {type(item).__name__} does not match expected type {item_type[0].__name__}.")

                    # If target_type is a subclass of set, return an instance of target_type
                    if self._is_subclass_of_generic(target_type, set):
                        return target_type(instantiated_items)

                    return instantiated_items

                # Handle deques
                if origin is deque or (isinstance(origin, type) and issubclass(origin, set)):
                   item_type = get_args(target_type)[0] if get_args(target_type) else Any
                   return deque(self.instantiate(item, item_type) for item in data)

                if origin is frozenset or (isinstance(origin, type) and issubclass(origin, frozenset)):
                   item_type = get_args(target_type)[0] if get_args(target_type) else Any
                   return frozenset(self.instantiate(item, item_type) for item in data)

            except TypeError as e:
                print(e)
                raise TypeError(f"Failed to instantiate {target_type} from list. {e}")

        # If none of the above, return the data as-is
        return data



import openai
import os
from pydantic import BaseModel
import sys
import wikipedia

sys.path.append("../../src")
import tanuki

from dotenv import load_dotenv
load_dotenv()
openai.api_key = os.getenv("OPENAI_API_KEY")


# -- Simple Summary Example --
@tanuki.patch
def explain_simple(summary: str) -> str:
    """Explain the summary in simple terms."""


def ask_wikipedia(topic: str) -> str:
    summary = wikipedia.summary(topic)
    return explain_simple(summary)


def simplify_example(topic: str) -> None:
    print("Wikipedia Summary:\n")
    print(wikipedia.summary(topic))
    print("Simplify Summary:\n")
    print(ask_wikipedia(topic))


# -- Classify Example --
class Dinosaur(BaseModel):
    name: str
    nickname: str
    height: int
    weight: int


@tanuki.patch
def dinosaur_classifer(summary: str) -> Dinosaur:
    """Convert the input summary into a Dinosaur object."""

def dinopedia(dinosaur: str) -> Dinosaur:
    summary = wikipedia.summary(dinosaur)
    print(summary)
    print(dinosaur_classifer(summary))


if __name__ == "__main__":
    # simplify_example("Nuclear fission")

    # dino = "Tyrannosaurus"
    # dino = "Triceratops"
    # dino = "Stegosaurus"
    dino = "Velociraptor"
    # dino = "Spinosaurus"
    dinopedia(dino)



# import abstract base class
from abc import ABC, abstractmethod


class LLM_API(ABC):
    def __init__(self) -> None:
        pass
        
    @abstractmethod  
    def generate(self, model, system_message, prompt, **kwargs):
        """
        The main generation function, given the args, kwargs, function_modeler, function description and model type, generate a response and check if the datapoint can be saved to the finetune dataset
        """
        pass


# import abstract base class
from abc import ABC, abstractmethod
from typing import List

from tanuki.models.embedding import Embedding


class Embedding_API(ABC):
    def __init__(self) -> None:
        pass
        
    @abstractmethod
    def embed(self, texts: List[str], model: str = None, **kwargs) -> List[Embedding]:
        """
        The main embedding function, given the model and prompt, return a vector representation
        """
        pass


from dataclasses import dataclass
from datetime import datetime

from pydantic import Field
import tanuki
import unittest
from typing import Literal, Optional, List



@dataclass(frozen=True)
class ActionItem:
    goal: str = Field(..., description="The goal of the action item")
    deadline: datetime = Field(..., description="The deadline of the action item")
    completed: bool = Field(False, description="Whether the action item is completed")


class TestActionItems(unittest.TestCase):


    @tanuki.patch
    def create_action_items(self, input: str) -> List[ActionItem]:
        """
        Create action items from a bit of text.
        """
        pass

    @tanuki.align
    def test_align_create_action_items_of_fixed_length(self):
        """We can test the function as normal using Pytest or Unittest"""

        speech = "Heya, I need you to do three things for me. " \
                 "First, I need you to get me a coffee by 3pm. " \
                 "Second, I need you to get me a cheese by 4pm. " \
                 "Third, I need you to get me a chocolate by 5pm on Tuesday."
        output = self.create_action_items(speech)
        assert len(output) == 3




if __name__ == '__main__':
    unittest.main()



import hashlib
from dataclasses import dataclass
from typing import Dict, Optional, Literal

from tanuki.models.function_type import FunctionType
from tanuki.utils import json_dumps

@dataclass(frozen=True)
class FunctionDescription:
    name: str
    docstring: str
    input_type_hints: Dict[str, type]
    input_class_definitions: Dict[str, str]
    output_type_hint: type
    output_class_definition: Optional[str]
    type: FunctionType = FunctionType.SYMBOLIC

    def __hash__(self, purpose: str = "general"):
        if purpose == "general": 
            json_encoded = json_dumps(self).encode('utf-8')
            h = hashlib.md5(json_encoded).hexdigest()
            return str(h)
        if purpose == "finetune": 
            json_encoded = json_dumps(self).encode('utf-8')
            h = hashlib.shake_256(json_encoded).hexdigest(8)
            return str(h)


import dis
from dotenv import load_dotenv
load_dotenv()
import tanuki


@tanuki.patch
def semantic_sql(input: str, schema: str) -> str:
    """
    Convert the input into a valid SQL query. Assume that the database has a table named "user" with
    columns "id", "organization_id", "name", and "email_notify"
    """

@tanuki.patch
def semantic_sql_var(input: str, table_name: str = None, columns=[]) -> str:
    """
    Convert the input into a valid SQL query. Assume that the database has a table named "user" with
    columns "id", "organization_id", "name", and "email_notify"
    """


@tanuki.align
def test_semantic_sql_var():
    output = semantic_sql_var("list the names of all users in our user database", table_name="user", columns=["id", "organization_id", "name", "email_notify"])
    assert output == "SELECT name FROM user"

@tanuki.align
def test_semantic_sql():
    output = semantic_sql(input = "list the names of all users in our user database", schema = """table_name=user, columns=["id", "organization_id", "name", "email_notify"]""")
    assert output == "SELECT name FROM user"


if __name__ == '__main__':
    test_semantic_sql_var()
    #test_semantic_sql()
    print("Everything passed")


import openai
import os
from dotenv import load_dotenv
from pydantic import BaseModel
from typing import List, Optional

load_dotenv()

import tanuki
from utils import scrape_url


openai.api_key = os.getenv("OPENAI_API_KEY")


class Car(BaseModel):
    price: float
    mpg: str
    seating: int
    horsepower: int
    weight: int
    fuel_size: float
    warranty_basic: str
    warranty_powertrain: str
    warranty_roadside: str


@tanuki.patch
def extract_car(content: str) -> Optional[Car]:
    """
    Examine the content string and extract the car details for the price, miles per gallon, seating, horsepower,
    weight, fuel tank size, and warranty.
    """


if __name__ == '__main__':

    # Web scrape the url and extract the car information
    # url = "https://www.cars.com/research/ford-mustang-2024/"
    url = "https://www.cars.com/research/mazda-cx_90-2024/"
    contents = scrape_url(url=url)
    print(contents)

    # Process the cocktail block using Tanuki
    car = extract_car(contents[0])
    print(car)



import openai
import os
from dotenv import load_dotenv
from pydantic import BaseModel
from typing import Optional

load_dotenv()

import tanuki
from utils import scrape_url


openai.api_key = os.getenv("OPENAI_API_KEY")


class Country(BaseModel):
    name: str
    capital: str
    population: int
    area: float


@tanuki.patch
def extract_country(content: str) -> Optional[Country]:
    """
    Examine the content string and extract the country information pertaining to it's
    name, capital, population, and area.
    """


@tanuki.align
def align_extract_country() -> None:
    print("Aligning...")
    country = "\n\n\n                            U.S. Virgin Islands\n                        \n\nCapital: Charlotte Amalie\nPopulation: 108708\nArea (km2): 352.0\n\n"
    assert extract_country(country) == Country(
        name="U.S. Virgin Islands",
        capital="Charlotte Amalie",
        population=108708,
        area=352.0,
    )


if __name__ == '__main__':

    # Align the function
    align_extract_country()

    # Web scrape the url and extract the list of countries
    url = "https://www.scrapethissite.com/pages/simple/"
    contents = scrape_url(url=url, class_name="country")

    # Process the country blocks using Tanuki (only sampling a couple for demo purposes)
    countries = []
    for content in contents[10:12]:
        countries.append(extract_country(content))
    print(countries)



import os
import unittest
from typing import Literal, Optional, List, Union
from pydantic import BaseModel, Field
import openai
from dotenv import load_dotenv

import tanuki

load_dotenv()
from tanuki.register import Register
openai.api_key = os.getenv("OPENAI_API_KEY")

class Person(BaseModel):
    age: int = Field(..., ge=0, le=155)
    name: str
    favourite_colours: List[str]



@tanuki.patch
def summarise_list_generic(input: str) -> List[str]:
    """
    Summarise the input into multiple sentences in a list
    """
@tanuki.patch
def summarise_list_typing(input: str) -> list[str]:
    """
    Summarise the input into multiple sentences in a list
    """
@tanuki.patch
def summarise_list_pydantic(input: str) -> List[Person]:
    """
    Create a list of Personas
    """
@tanuki.patch
def summarise_list_dict(input: str) -> List[dict]:
    """
    Create a list of dict personas
    """

@tanuki.patch
def summarise_list_int(input: str) -> List[int]:
    """
    Extract the integers
    """

@tanuki.patch
def summarise_list_Union(input: str) -> List[Union[int, float]]:
    """
    Extract the numbers
    """

@tanuki.align
def align_list_generic():

    assert summarise_list_generic("Thats awesome. Thats cool") == ["Thats awesome", "Thats cool"]
    assert summarise_list_generic("Thats neat. Thats ok") == ["Thats neat", "Thats ok"]

    assert summarise_list_generic(input = "Thats awesome. Thats cool") == ["Thats awesome", "Thats cool"]
    assert summarise_list_generic(input = "Thats neat. Thats ok") == ["Thats neat", "Thats ok"]

@tanuki.align
def align_list_typing():
    assert summarise_list_typing("Thats awesome. Thats cool") == ["Thats awesome", "Thats cool"]
    assert summarise_list_typing("Thats neat. Thats ok") == ["Thats neat", "Thats ok"]

    assert summarise_list_typing(input = "Thats awesome. Thats cool") == ["Thats awesome", "Thats cool"]
    assert summarise_list_typing(input = "Thats neat. Thats ok") == ["Thats neat", "Thats ok"]


@tanuki.align
def align_list_pydantic():

    person_str = "First person - Name Jeff, age 25, favourite colours Red and Blue. Second person - Name Thomas, age 33, favourite colours Green and Gray"
    output = [Person(name="Jeff", age=25, favourite_colours=["Red", "Blue"]), Person(name="Thomas", age=33, favourite_colours=["Green", "Gray"])]
    assert summarise_list_pydantic(person_str) == output
    assert summarise_list_pydantic(input = person_str) == output


@tanuki.align
def align_list_dict():
    person_str = "First person - Name Jeff, age 25, favourite colours Red and Blue. Second person - Name Thomas, age 33, favourite colours Green and Gray"
    output = [{"name": "Jeff", "age": 25, "favourite_colours": ["Red", "Blue"]}, {"name": "Thomas", "age": 33, "favourite_colours": ["Green", "Gray"]}]
    assert summarise_list_dict(person_str) == output
    assert summarise_list_dict(input = person_str) == output


@tanuki.align
def align_list_int():

    input_1 = "1 and 2"
    input_2 = "1, 2 and 3"

    assert summarise_list_int(input_1) == [1, 2]
    assert summarise_list_int(input_2) == [1, 2, 3]

    assert summarise_list_int(input = input_1) == [1, 2]
    assert summarise_list_int(input = input_2) == [1, 2, 3]

@tanuki.align
def align_list_Union():
    input_1 = "1 and 2"
    input_2 = "1.0, 2.0 and 3.0"

    assert summarise_list_Union(input_1) == [1, 2]
    assert summarise_list_Union(input_2) ==[1.0, 2.0, 3.0]

    assert summarise_list_Union(input = input_1) == [1, 2]
    assert summarise_list_Union(input = input_2) == [1.0, 2.0, 3.0]

def test_list():
    # This tests all the list aligns
    # can be called by pytest or unittest
    align_list_generic()
    align_list_typing()
    align_list_pydantic()
    align_list_dict()
    align_list_int()
    align_list_Union()
    print("All list aligns passed!")

@tanuki.patch
def summarise_str(input: str) -> str:
    """
    Summarise the input into 1 sentence
    """

@tanuki.patch
def summarise_pydantic(input: str) -> Person:
    """
    Create the persona
    """
@tanuki.patch
def summarise_dict(input: str) -> dict:
    """
    Create the persona
    """

@tanuki.patch
def summarise_int(input: str) -> int:
    """
    Extract the integer
    """

@tanuki.patch
def summarise_Union(input: str) -> Union[int, float]:
    """
    Extract the number
    """

@tanuki.align
def align_string():
    assert summarise_str("Thats awesome. Thats cool") == 'They found it awesome and cool'
    assert summarise_str("Thats neat. Thats ok") == 'They found it neat and ok'

    assert summarise_str(input = "Thats awesome. Thats cool") == 'They found it awesome and cool'
    assert summarise_str(input = "Thats neat. Thats OK") == 'They found it neat and ok'


@tanuki.align
def align_pydantic():
    input_str = "Name Jeff, age 25, favourite colours Red and Blue"
    person = Person(name="Jeff", age=25, favourite_colours=["Red", "Blue"])
    assert summarise_pydantic(input_str) == person

    assert summarise_pydantic(input = input_str) == person

@tanuki.align
def align_dict():
    input_str = "Name Jeff, age 25, favourite colours Red and Blue"
    output = {"name": "Jeff", "age": 25, "favourite_colours": ["Red", "Blue"]}
    assert summarise_dict(input_str) == output

    assert summarise_dict(input = input_str) == output


@tanuki.align
def align_list_int():
    input_str = "This is number 1"
    assert summarise_int(input_str) == 1
    assert summarise_int(input = input_str) == 1

@tanuki.align
def align_list_Union():
    input_str_1 = "This is number 1"
    input_str_2 = "This is number 2.0"
    assert summarise_Union(input_str_1) == 1
    assert summarise_Union(input_str_2) == 2.0

    assert summarise_Union(input = input_str_1) == 1
    assert summarise_Union(input = input_str_2) == 2.0

def test_single():
    # This tests all the single aligns
    # can be called by pytest or unittest
    align_string()
    align_pydantic()
    align_dict()
    align_list_int()
    align_list_Union()
    print("All single aligns passed!")

def _parse_examples(test_func):
    # check that all the examples are correctly readable
    function_description = Register.load_function_description(test_func)
    function_modeler = tanuki.function_modeler
    align_examples = function_modeler.get_symbolic_alignments(function_description.__hash__())
    examples = "\n".join([f"Inputs:\nArgs: {align['args']}\nKwargs: {align['kwargs']}\nOutput: {align['output']}" for align in align_examples])

def test_parse_align_datasets():
    # Test that all the examples that are aligned are correctly parsable into the prompt format we have defined
    _parse_examples(summarise_list_generic)
    _parse_examples(summarise_list_typing)
    _parse_examples(summarise_list_pydantic)
    _parse_examples(summarise_list_dict)
    _parse_examples(summarise_list_int)
    _parse_examples(summarise_list_Union)
    _parse_examples(summarise_str)
    _parse_examples(summarise_pydantic)
    _parse_examples(summarise_dict)
    _parse_examples(summarise_int)
    _parse_examples(summarise_Union)
    print("All examples parsed correctly!")



from typing import List

import openai
import time
# import abstract base class
from tanuki.language_models.embedding_api_abc import Embedding_API
from tanuki.language_models.llm_api_abc import LLM_API
import os

OPENAI_URL = "https://api.openai.com/v1/chat/completions"
import requests


class OpenAI_API(LLM_API, Embedding_API):
    def __init__(self) -> None:
        # initialise the abstract base class
        super().__init__()
        self.api_key = os.getenv("OPENAI_API_KEY")

    def embed(self, texts: List[str], model="text-similarity-babbage-001", **kwargs):
        """
        Generate embeddings for the provided texts using the specified OpenAI model.
        Lightweight wrapper over the OpenAI client.

        :param texts: A list of texts to embed.
        :param model: The model to use for embeddings.
        :return: A list of embeddings.
        """
        self.check_api_key()

        try:
            response = openai.Embedding.create(
                input=texts,
                model=model,
                **kwargs
            )
            return response['data']
        except Exception as e:
            print(f"An error occurred: {e}")
            return None

    def generate(self, model, system_message, prompt, **kwargs):
        """
        The main generation function, given the args, kwargs, function_modeler, function description and model type, generate a response and check if the datapoint can be saved to the finetune dataset
        """

        self.check_api_key()

        temperature = kwargs.get("temperature", 0)
        top_p = kwargs.get("top_p", 1)
        frequency_penalty = kwargs.get("frequency_penalty", 0)
        presence_penalty = kwargs.get("presence_penalty", 0)
        params = {
            "model": model,
            "temperature": temperature,
            "max_tokens": 512,
            "top_p": top_p,
            "frequency_penalty": frequency_penalty,
            "presence_penalty": presence_penalty,
        }
        messages = [
            {
                "role": "system",
                "content": system_message
            },
            {
                "role": "user",
                "content": prompt
            }
        ]
        params["messages"] = messages

        counter = 0
        choice = None
        # initiate response so exception logic doesnt error out when checking for error in response
        response = {}
        while counter <= 5:
            try:
                openai_headers = {
                    "Authorization": f"Bearer {self.api_key}",
                    "Content-Type": "application/json",
                }
                response = requests.post(
                    OPENAI_URL, headers=openai_headers, json=params, timeout=50
                )
                response = response.json()
                choice = response["choices"][0]["message"]["content"].strip("'")
                break
            except Exception as e:
                if ("error" in response and
                        "code" in response["error"] and
                        response["error"]["code"] == 'invalid_api_key'):
                    raise Exception(f"The supplied OpenAI API key {self.api_key} is invalid")
                if counter == 5:
                    raise Exception(f"OpenAI API failed to generate a response: {e}")
                counter += 1
                time.sleep(2 ** counter)
                continue

        if not choice:
            raise Exception("OpenAI API failed to generate a response")

        return choice

    def check_api_key(self):
        # check if api key is not none
        if self.api_key is None:
            # try to get the api key from the environment, maybe it has been set later
            self.api_key = os.getenv("OPENAI_API_KEY")
            if self.api_key is None:
                raise ValueError("OpenAI API key is not set")



import inspect
import os

import openai
from dotenv import load_dotenv

import tanuki
load_dotenv()
openai.api_key = os.getenv("OPENAI_API_KEY")

@tanuki.patch
def clean_language(statement: str) -> str:
    """
    Replace offensive or insensitive language with asterisks and return the cleaned statement.
    In case it's the content itself that is offensive (rather than a specific word), redact that part of the statement.
    """

@tanuki.align
def test_clean_language():
    """We can test the function as normal using Pytest or Unittest"""
    assert clean_language("I think you should kill yourself") == 'I think you should [*redacted*]'
    assert clean_language("I want you to take a flying fuck at a rolling donut") == "I want you to take a [*redacted*]"


if __name__ == '__main__':
    test_clean_language()
    print(clean_language("I think you should jump off a bridge"))
    print(clean_language("Jesus fucking christ, how can you say that?"))


import os
from enum import Enum
from typing import Literal, Union, Optional, Dict

from appdirs import user_data_dir

from tanuki.constants import *
from tanuki.persistence.filter.bloom_interface import IBloomFilterPersistence
from tanuki.persistence.filter.filesystem_bloom import BloomFilterFileSystemDriver
from tanuki.trackers.abc_buffered_logger import ABCBufferedLogger


class FilesystemBufferedLogger(ABCBufferedLogger):
    """
    A class that handles the reading and writing of patch invocations and align statements.
    It includes the logic for a bloom filter, to ensure that we only store unique invocations.
    """

    def __init__(self, name, level=15):
        self.log_directory = self._get_log_directory()
        super().__init__(name, level)

    def get_bloom_filter_persistence(self) -> IBloomFilterPersistence:
        """
        Get an instance of the bloom filter persistence provider. Typically this will be a file system provider.
        :return: A persistence provider
        """
        return BloomFilterFileSystemDriver(log_directory=self.log_directory)

    def get_patch_location_for_function(self, func_hash, extension: Union[
        ALIGN_FILE_EXTENSION_TYPE, PATCH_FILE_EXTENSION_TYPE] = "") -> str:

        """
        Get the local location of the function patch file.
        :param func_hash: The representation of the function
        :param extension: Whether this is a patch or an alignment
        :return:
        """
        return os.path.join(self.log_directory, func_hash + extension)

    def ensure_persistence_location_exists(self) -> None:
        """
        Ensure that the location on the filesystem we will be writing to actually exists. If not, create it.
        """
        log_directory = self.log_directory
        # Create the folder if it doesn't exist
        if not os.path.exists(log_directory):
            os.makedirs(log_directory)

    def does_object_exist(self, path: str) -> bool:
        """
        Check to see if a path exists on the filesystem.
        :param path:
        :return:
        """
        return os.path.exists(path)

    def _get_log_directory(self) -> str:
        """
        Find a location on the filesystem to write our logs to.
        :return:
        """
        filename = "functions"

        # If explicitly defined
        env_dir = os.getenv(ENVVAR)
        if env_dir and os.path.isdir(env_dir):
            return os.path.join(env_dir, filename)

        # If installed as a library
        library_dir = os.path.join(user_data_dir(LIB_NAME), filename)
        if os.path.isdir(library_dir) or not os.path.exists(library_dir):
            return library_dir

        # If installed in a project that contains a git repo - place it in the same folder as the git repo
        current_dir = os.getcwd()
        while current_dir != os.path.root:
            if ".git" in os.listdir(current_dir):
                return os.path.join(current_dir, filename)
            current_dir = os.path.dirname(current_dir)

        return os.path.join(os.getcwd(), filename)

    def load_dataset(self, dataset_type, func_hash, return_type="both") -> Optional[int]:
        """
        Get the size of the dataset for a function hash
        """
        log_directory = self._get_log_directory()
        dataset_type_map = {"alignments": ALIGN_FILE_EXTENSION,
                            "positive": POSITIVE_FILE_EXTENSION,
                            "negative": NEGATIVE_FILE_EXTENSION,
                            "patches": PATCH_FILE_EXTENSION}

        log_file_path = os.path.join(log_directory, func_hash + dataset_type_map[dataset_type])
        if not os.path.exists(log_file_path):
            if return_type == "both":
                return 0, None
            elif return_type == "dataset":
                return None
            elif return_type == "length":
                return 0
        try:
            with open(log_file_path, "rb") as f:
                dataset = f.read()
            dataset_string = repr(dataset)
            dataset_length = dataset_string.count("\\n") - dataset_string.count("\\\\n")
            if return_type == "both":
                return dataset_length, dataset
            elif return_type == "dataset":
                return dataset
            elif return_type == "length":
                return dataset_length
        except Exception as e:
            if return_type == "both":
                return 0, None
            elif return_type == "dataset":
                return None
            elif return_type == "length":
                return 0

    def load_existing_datasets(self) -> Dict[str, Dict[str, str]]:
        log_directory = self.log_directory
        dataset_lengths = {
            SYMBOLIC_ALIGNMENTS: {},
            POSITIVE_EMBEDDABLE_ALIGNMENTS: {},
            NEGATIVE_EMBEDDABLE_ALIGNMENTS: {},
            PATCHES: {},
        }

        try:
            if not os.path.exists(log_directory):
                os.makedirs(log_directory)
            # get all the files in the log directory
            files = os.listdir(log_directory)
            # discard all .json files
            files = [x for x in files if ".json" not in x]
        except Exception as e:
            return dataset_lengths

        for file in files:
            if ALIGN_FILE_EXTENSION not in file \
                    and PATCH_FILE_EXTENSION not in file \
                    and POSITIVE_FILE_EXTENSION not in file \
                    and NEGATIVE_FILE_EXTENSION not in file:
                continue
            elif ALIGN_FILE_EXTENSION in file:
                dataset_type = SYMBOLIC_ALIGNMENTS
            elif POSITIVE_FILE_EXTENSION in file:
                dataset_type = POSITIVE_EMBEDDABLE_ALIGNMENTS
            elif NEGATIVE_FILE_EXTENSION in file:
                dataset_type = NEGATIVE_EMBEDDABLE_ALIGNMENTS
            else:
                dataset_type = PATCHES
            func_hash = file.replace(ALIGN_FILE_EXTENSION, "").replace(PATCH_FILE_EXTENSION, "")
            dataset_lengths[dataset_type][func_hash] = -1
        return dataset_lengths

    def write(self, path: str, data: str, mode: Literal["w", "a", "a+b"] = "w") -> None:
        """
        Write data to a file
        """
        with open(path, mode) as f:
            f.write(data)

    def read(self, path: str) -> str:
        """
        Read data from a file
        """
        with open(path, "r") as f:
            return f.read()

    def get_hash_from_path(self, path) -> str:
        """
        Given a path with a hash, return only the hash
        :param path: The path to the file
        :return: The hash
        """
        return path.replace(PATCH_FILE_EXTENSION, ""). \
            replace(self.log_directory, ""). \
            lstrip("/"). \
            lstrip("\\")



import os

# Load in the environment variables
import dotenv
dotenv.load_dotenv()

# Initialize OpenAI
import openai
openai.api_key = os.getenv("OPENAI_API_KEY") # Your OpenAI API key

# import Tanuki
import tanuki
# import Pydantic to define the response type
from pydantic import Field, BaseModel

# Additional imports for the demo that will be used to fetch a random product from the Best Buy API
from random import randint
import requests


# First we define the desired response type using Pydantic
class ProductTags(BaseModel):
    category: str = Field(..., description="What category does this product fit into?")
    brand: str = Field(..., description="What brand is this product?")
    features: str = Field(..., description="What are the features of this product?")


# Next we define the function that will be used to generate the response with the Monkey Patch decorator
@tanuki.patch
def product_tagger(product_story: str) -> ProductTags:
    """
    based on the provided product story, extract the following aspects:
    - what category does this product fit into? (e.g., Sports & outdoors, Home Goods, Mens Clothing, Garden Tools, etc.)
    - what brand is this product? (e.g., Nike, Adidas, Apple, Samsung, etc.)
    - what are the features of this product? (e.g., Bluetooth, 4K, 5G, water-resistant, removable liner, etc.)
    If it is not possible to extract any of these aspects, return "unknown" for that aspect.
    """

# (OPTIONAL!) We use the align decorator to both test the function, and to teach the model about our expected output.
# This can include as many or as few assert statements as you like.
@tanuki.align
def test_product_tagger():
    """We can test the function as normal using Pytest or Unittest"""
    assert product_tagger("Product: Apple iPhone 12 Pro Max 5G 128GB - Pacific Blue (Verizon) Short Description: "
                          "iPhone 12 Pro Max. 5G to download movies on the fly and stream high-quality video.¹ "
                          "Beautifully bright 6.7-inch Super Retina XDR display.² Ceramic Shield with 4x better drop "
                          "performance.³ Incredible low-light photography with the best Pro camera system on an "
                          "iPhone, and 5x optical zoom range. Cinema-grade Dolby Vision video recording, editing, "
                          "and playback. Night mode portraits and next-level AR experiences with the LiDAR Scanner. "
                          "Powerful A14 Bionic chip. And new MagSafe accessories for easy attach and faster wireless "
                          "charging.⁴ For infinitely spectacular possibilities. Features: 6.7-inch Super Retina XDR "
                          "display Ceramic Shield, tougher than any smartphone glass A14 Bionic chip, the fastest "
                          "chip ever in a smartphone Pro camera system with 12MP Ultra Wide, Wide, and Telephoto "
                          "cameras; 5x optical zoom range; Night mode, Deep Fusion, Smart HDR 3, Apple ProRAW, "
                          "4K Dolby Vision HDR recording LiDAR Scanner for improved AR experiences, Night mode "
                          "portraits 12MP TrueDepth front camera with Night mode, 4K Dolby Vision HDR recording "
                          "Industry-leading IP68 water resistance Supports MagSafe accessories for easy attach and "
                          "faster wireless charging iOS with redesigned widgets on the Home screen, all-new App "
                          "Library, App Clips and more") == ProductTags(
        category='Electronics',
        brand='Apple',
        features='5G, 4K, Dolby Vision, LiDAR Scanner, Night mode, 12MP TrueDepth front camera, IP68 water '
                 'resistance, MagSafe accessories, iOS'
    )


# Finally we can run the function on a random product from the Best Buy API
if __name__ == '__main__':

    # get the Best Buy API key from the environment variables (this is set in .env file)
    bestbuy_api_key = os.getenv("BESTBUY_API")

    # get random number between 1 and 160000 (roughly the number of products in the Best Buy API)
    random_number = randint(1, 160000)

    # set the Best Buy API endpoint
    bestbuy_api = "https://api.bestbuy.com/v1/products"

    # fetch a random product from the API
    response = requests.get(f"{bestbuy_api}?apiKey={bestbuy_api_key}&pageSize=1&page={random_number}&format=json")
    response.raise_for_status()

    # get the product from the response (if no errors were raised)
    product = response.json()['products'][0]

    # Let's assemble the input for the product tagger
    product_story = f"""
    Product: {product['name']}
    Short Description: {product['shortDescription']}
    Description: {product['description']}
    Long Description: {product['longDescription']}
    """

    # (Optional) Run the test to ensure the function works as expected and to teach the model about the expected output
    test_product_tagger()

    # Run the function on the product story and show the results
    print("\nOriginal input:")
    print(product_story)
    print("\nGenerated output:")

    tagged_product = product_tagger(product_story)

    # we can safely print the model dump because we know its type is ProductTags
    print(tagged_product.model_dump())


import ast
import datetime
import pprint

from typing import Optional, List

from pydantic import BaseModel

import tanuki
from tanuki.assertion_visitor import AssertionVisitor, Or


def _parse(source):
    class TodoItem(BaseModel):
        deadline: Optional[datetime.datetime] = None
        goal: str
        people: List[str]

    @tanuki.patch
    def create_todolist_item(goal: str, people=[]) -> TodoItem:
        pass

    tree = ast.parse(source)
    _locals = locals()
    visitor = AssertionVisitor(locals(), patch_symbolic_funcs={"create_todolist_item": create_todolist_item})
    visitor.visit(tree)
    return visitor.mocks

def test_rl_equality():
    source = \
""" 
assert TodoItem(goal="Go to the store and buy some milk", people=["Me"]) == create_todolist_item("I would like to go to the shop and buy some milk")
"""

    mocks = _parse(source)

    assert len(mocks) == 1

def test_equality():

    source = \
""" 
assert create_todolist_item("I would like to go to the shop and buy some milk") == TodoItem(goal="Go to the store and buy some milk", people=["Me"])
"""

    mocks = _parse(source)

    assert len(mocks) == 1


def test_non_equality():
    """
    Should not create a mock, because we need input - output pairs.
    """
    source = \
"""
assert create_todolist_item("I need to go and visit George at 3pm tomorrow") != TodoItem(goal="Go and visit George", people=["Me"], deadline=datetime(2021, 1, 1, 14, 0))
"""

    mocks = _parse(source)

    assert len(mocks) == 0


def test_none():
    source = \
"""
assert create_todolist_item("Invalid input") is None
"""
    mocks = _parse(source)

    assert len(mocks) == 1

def test_not_none():
    """
    This should also not result in a mock, because nEq is insufficient to map input to output.
    :return:
    """
    source = \
"""
assert create_todolist_item("I would like to go to the shop and buy some tomatoes") is not None
"""
    mocks = _parse(source)

    assert len(mocks) == 0


def test_inclusion():
    """
    TODO: Review whether it is a good idea to output a list of valid outputs. This may occlude legitimate list outputs.
    Perhaps we want to create a special list-like object that can be used to represent a list of valid outputs.
    :return:
    """
    source = \
"""
import datetime
valid_outputs = [
        TodoItem(goal="Go to the store and buy some milk", people=["Me"]),
        TodoItem(goal="Go and visit Jeff", people=["Me"], deadline=datetime.datetime(2021, 1, 1, 15, 0))
    ]
assert create_todolist_item("I would like to go to the shop and buy some cheese") in valid_outputs
"""
    mocks = _parse(source)

    assert len(list(mocks.values())[0]) == 2


def test_exclusion():
    source = \
"""
import datetime
invalid_outputs = [
        TodoItem(goal="Go to the gym", people=["Me"]),
        TodoItem(goal="Go and visit Sarah", people=["Me"], deadline=datetime(2021, 1, 1, 16, 0))
    ]
assert create_todolist_item("I would like to go to the terrace and buy some milk") not in invalid_outputs
"""
    mocks = _parse(source)

    assert len(mocks) == 0


def test_tuple_iteration():
    source = \
"""
import datetime
inputs = ["I would like to go to the store and buy some milk", "I need to go and visit Jeff at 3pm tomorrow"]
outputs = [TodoItem(goal="Go to the store and buy some milk", people=["Me"]), TodoItem(goal="Go and visit Jeff", people=["Me"], deadline=datetime(2021, 1, 1, 15, 0))]
for input, output in zip(inputs, outputs):
    assert create_todolist_item(input) == output
"""
    mocks = _parse(source)

    assert len(mocks) == 2

def test_iteration():
    source = \
"""
import datetime
inputs = ["I would like to go to the store and buy some milk", "I need to go and visit Jeff at 3pm tomorrow"]
outputs = [TodoItem(goal="Go to the store and buy some milk", people=["Me"]), TodoItem(goal="Go and visit Jeff", people=["Me"], deadline=datetime(2021, 1, 1, 15, 0))]
for input in inputs:
    assert create_todolist_item(input) in outputs
"""
    mocks = _parse(source)

    assert len(mocks) == 2

    assert len(list(mocks.values())[0]) == 2
    assert len(list(mocks.values())[1]) == 2

    assert isinstance(list(mocks.values())[0], Or)

if __name__ == "__main__":
    test_equality()
    test_iteration()
    test_non_equality()
    test_none()
    test_not_none()
    test_inclusion()
    test_exclusion()
    test_tuple_iteration()
    test_rl_equality()




import os
import unittest
from typing import Literal, Optional
import openai
from dotenv import load_dotenv

import tanuki

load_dotenv()

openai.api_key = os.getenv("OPENAI_API_KEY")


class TestClassifySentiment(unittest.TestCase):

    @tanuki.patch
    def classify_sentiment_2(self, input: str, input_2: str) -> Optional[Literal['Good', 'Bad']]:
        """
        Determine if the inputs are positive or negative sentiment, or None
        """


    @tanuki.patch
    def classify_sentiment(self, input: str) -> Optional[Literal['Good', 'Bad']]:
        """
        Determine if the input is positive or negative sentiment
        """


    @tanuki.align
    def test_align_classify_sentiment(self):
        """We can test the function as normal using Pytest or Unittest"""

        i_love_you = "I love you"
        print(self.classify_sentiment_2(i_love_you, "I love woo"))
        assert self.classify_sentiment_2(i_love_you, "I love woo") == 'Good'

        assert self.classify_sentiment("I love you") == 'Good'

        assert self.classify_sentiment("I hate you") == 'Bad'

        assert not self.classify_sentiment("Wednesdays are in the middle of the week")


if __name__ == '__main__':
    unittest.main()


import os
import unittest
from unittest.mock import patch, Mock

from dotenv import load_dotenv

from tanuki.language_models.openai_api import OpenAI_API

load_dotenv()


class TestOpenAI_API(unittest.TestCase):

    @patch('os.getenv')
    def test_missing_api_key(self, mock_getenv):
        mock_getenv.return_value = None
        api = OpenAI_API()
        with self.assertRaises(ValueError):
            api.generate("model_name", "system_message", "prompt")

    @patch('requests.post')
    @patch('os.getenv')
    def test_invalid_api_key(self, mock_getenv, mock_post):
        mock_getenv.return_value = "invalid_key"
        mock_response = Mock()
        mock_response.json.return_value = {
            "error": {"code": 'invalid_api_key'}
        }
        mock_post.return_value = mock_response

        api = OpenAI_API()
        with self.assertRaises(Exception) as context:
            api.generate("model_name", "system_message", "prompt")
        self.assertIn("invalid", str(context.exception))

    @patch('requests.post')
    @patch('os.getenv')
    def test_successful_generation(self, mock_getenv, mock_post):
        mock_getenv.return_value = os.getenv("OPENAI_API_KEY")
        mock_response = Mock()
        mock_response.json.return_value = {
            "choices": [{"message": {"content": "Generated response"}}]
        }
        mock_post.return_value = mock_response

        api = OpenAI_API()
        result = api.generate("model_name", "system_message", "prompt")
        self.assertEqual(result, "Generated response")


if __name__ == '__main__':
    unittest.main()


from typing import List

from tanuki.function_modeler import FunctionModeler
from tanuki.language_models.language_model_manager import LanguageModelManager
from tanuki.register import Register
from tanuki.trackers.filesystem_buffered_logger import FilesystemBufferedLogger


def dummy_func(input: str) -> List[str]:
    """
    Below you will find an article with stocks analysis. Bring out the stock symbols of companies who are expected to go up or have positive sentiment
    """

def initiate_test(func_modeler, function_description):
    func_hash = function_description.__hash__()
    # initiate the config
    _ = func_modeler.load_function_config(func_hash, function_description)
    for keys, values in func_modeler.function_configs.items():
        if func_hash in keys:
            values["distilled_model"] = "test_ft_1"
            values["teacher_models"] = ["gpt-4","gpt-4-32k"] # model and its token limit]
    func_modeler._update_config_file(func_hash)

def test_token_counter_finetunable():
    args = (0,)
    kwargs = {}
    function_description = Register.load_function_description(dummy_func)
    logger = FilesystemBufferedLogger("test")

    func_modeler = FunctionModeler(logger)
    lang_model = LanguageModelManager(func_modeler)

    initiate_test(func_modeler, function_description)

    prompt, distilled_model, suitable_for_distillation, is_distilled_model = lang_model.get_generation_case(args, kwargs, function_description)
    assert suitable_for_distillation
    assert is_distilled_model
    assert distilled_model == "test_ft_1"

def test_token_counter_non_finetunable_1():
    input = "(" * 6997
    args = (input,)
    kwargs = {}
    function_description = Register.load_function_description(dummy_func)
    logger = FilesystemBufferedLogger("test")
    func_modeler = FunctionModeler(logger)
    lang_model = LanguageModelManager(func_modeler)
    initiate_test(func_modeler, function_description)

    prompt, distilled_model, suitable_for_distillation, is_distilled_model = lang_model.get_generation_case(args, kwargs, function_description)
    assert not suitable_for_distillation
    assert not is_distilled_model
    assert distilled_model == "gpt-4"

def test_token_counter_non_finetunable_2():
    input = "(" * 7700
    args = (input,)
    kwargs = {}
    function_description = Register.load_function_description(dummy_func)
    logger = FilesystemBufferedLogger("test")
    func_modeler = FunctionModeler(logger)
    lang_model = LanguageModelManager(func_modeler)
    initiate_test(func_modeler, function_description)

    prompt, distilled_model, suitable_for_distillation, is_distilled_model = lang_model.get_generation_case(args, kwargs, function_description)
    assert not suitable_for_distillation
    assert not is_distilled_model
    assert distilled_model == "gpt-4-32k"

def test_error_raise():
    input = "(" * 32000
    args = (input,)
    kwargs = {}
    function_description = Register.load_function_description(dummy_func)
    #func_hash = function_description.__hash__()
    logger = FilesystemBufferedLogger("test")
    func_modeler = FunctionModeler(logger)
    lang_model = LanguageModelManager(func_modeler)
    initiate_test(func_modeler, function_description)
    error = False
    try:
        prompt, distilled_model, suitable_for_distillation, is_distilled_model = lang_model.get_generation_case(args, kwargs, function_description)
    except ValueError:
        error = True
    assert error

if __name__ == '__main__':
    test_token_counter_finetunable()
    test_token_counter_non_finetunable_1()
    test_token_counter_non_finetunable_2()
    test_error_raise()


import os
import unittest
from typing import Literal, Optional, List, Union
from pydantic import BaseModel, Field
import openai
from dotenv import load_dotenv

import tanuki

load_dotenv()
from tanuki.register import Register
openai.api_key = os.getenv("OPENAI_API_KEY")

class Person(BaseModel):
    age: int = Field(..., ge=0, le=155)
    name: str
    favourite_colours: List[str]



@tanuki.patch
def summarise_list_generic(input: List[str]) -> str:
    """
    Summarise the input
    """
@tanuki.patch
def summarise_list_typing(input: list[str]) -> str:
    """
    Summarise the input
    """
@tanuki.patch
def summarise_list_pydantic(input: List[Person]) -> str:
    """
    Summarise the input
    """
@tanuki.patch
def summarise_list_dict(input: List[dict]) -> str:
    """
    Summarise the input
    """

@tanuki.patch
def summarise_list_int(input: List[int]) -> str:
    """
    Summarise the input
    """

@tanuki.patch
def summarise_list_Union(input: List[Union[int, float]]) -> str:
    """
    Summarise the input
    """

@tanuki.align
def align_list_generic():
    assert summarise_list_generic(["Thats awesome", "Thats cool"]) == 'They found it awesome and cool'
    assert summarise_list_generic(["Thats neat", "Thats ok"]) == 'They found it neat and ok'

    assert summarise_list_generic(input = ["Thats awesome", "Thats cool"]) == 'They found it awesome and cool'
    assert summarise_list_generic(input = ["Thats neat", "Thats ok"]) == 'They found it neat and ok'

@tanuki.align
def align_list_typing():
    assert summarise_list_typing(["Thats awesome", "Thats cool"]) == 'They found it awesome and cool'
    assert summarise_list_typing(["Thats neat", "Thats ok"]) == 'They found it neat and ok'

    assert summarise_list_typing(input = ["Thats awesome", "Thats cool"]) == 'They found it awesome and cool'
    assert summarise_list_typing(input = ["Thats neat", "Thats ok"]) == 'They found it neat and ok'


@tanuki.align
def align_list_pydantic():
    person_1 = Person(name="Jeff", age=25, favourite_colours=["Red", "Blue"])
    person_2 = Person(name="Thomas", age=33, favourite_colours=["Green", "Gray"])
    input = [person_1, person_2]
    assert summarise_list_pydantic(input) == "Jeff is 25 years old and likes Red and Blue. Thomas is 33 years old and likes Green and Gray."
    assert summarise_list_pydantic(input = input) == "Jeff is 25 years old and likes Red and Blue. Thomas is 33 years old and likes Green and Gray."


@tanuki.align
def align_list_dict():
    input = [{"name": "Jeff", "age": 25, "favourite_colours": ["Red", "Blue"]}, {"name": "Thomas", "age": 33, "favourite_colours": ["Green", "Gray"]}]
    assert summarise_list_dict(input) == "Jeff is 25 years old and likes Red and Blue. Thomas is 33 years old and likes Green and Gray."
    assert summarise_list_dict(input = input) == "Jeff is 25 years old and likes Red and Blue. Thomas is 33 years old and likes Green and Gray."


@tanuki.align
def align_list_int():
    assert summarise_list_int([1, 2]) == "The sum of the list is 3"
    assert summarise_list_int([1, 2, 3]) == "The sum of the list is 6"

    assert summarise_list_int(input = [1, 2]) == "The sum of the list is 3"
    assert summarise_list_int(input = [1, 2, 3]) == "The sum of the list is 6"

@tanuki.align
def align_list_Union():
    assert summarise_list_Union([1, 2]) == "The sum of the list is 3"
    assert summarise_list_Union([1.0, 2.0, 3.0]) == "The sum of the list is 6.0"

    assert summarise_list_Union(input = [1, 2]) == "The sum of the list is 3"
    assert summarise_list_Union(input = [1.0, 2.0, 3.0]) == "The sum of the list is 6.0"

def test_list():
    # This tests all the list aligns
    # can be called by pytest or unittest
    align_list_generic()
    align_list_typing()
    align_list_pydantic()
    align_list_dict()
    align_list_int()
    align_list_Union()
    print("All list aligns passed!")


@tanuki.patch
def summarise_str(input: str) -> str:
    """
    Summarise the input
    """

@tanuki.patch
def summarise_pydantic(input: Person) -> str:
    """
    Summarise the input
    """
@tanuki.patch
def summarise_dict(input: dict) -> str:
    """
    Summarise the input
    """

@tanuki.patch
def summarise_int(input: int) -> str:
    """
    Summarise the input
    """

@tanuki.patch
def summarise_Union(input: Union[int, float]) -> str:
    """
    Summarise the input
    """

@tanuki.align
def align_string():
    assert summarise_str("Thats awesome") == 'They found it awesome and cool'
    assert summarise_str("Thats neat") == 'They found it neat and ok'

    assert summarise_str(input = "Thats awesome") == 'They found it awesome and cool'
    assert summarise_str(input = "Thats neat") == 'They found it neat and ok'


@tanuki.align
def align_pydantic():
    person_1 = Person(name="Jeff", age=25, favourite_colours=["Red", "Blue"])
    input = person_1
    assert summarise_pydantic(input) == "Jeff is 25 years old and likes Red and Blue."

    assert summarise_pydantic(input = input) == "Jeff is 25 years old and likes Red and Blue."

@tanuki.align
def align_dict():
    input = {"name": "Jeff", "age": 25, "favourite_colours": ["Red", "Blue"]}
    assert summarise_dict(input) == "Jeff is 25 years old and likes Red and Blue."

    assert summarise_dict(input = input) == "Jeff is 25 years old and likes Red and Blue."


@tanuki.align
def align_list_int():
    assert summarise_int(1) == "This is number 1"
    assert summarise_int(input = 1) == "This is number 1"

@tanuki.align
def align_list_Union():
    assert summarise_Union(1) == "This is number 1"
    assert summarise_Union(2.0) == "This is number 2"

    assert summarise_Union(input = 1) == "This is number 1"
    assert summarise_Union(input = 2.0) == "This is number 2"

def test_single():
    # This tests all the single aligns
    # can be called by pytest or unittest
    align_string()
    align_pydantic()
    align_dict()
    align_list_int()
    align_list_Union()
    print("All single aligns passed!")

def _parse_examples(test_func):
    # check that all the examples are correctly readable
    function_description = Register.load_function_description(test_func)
    function_modeler = tanuki.function_modeler
    align_examples = function_modeler.get_symbolic_alignments(function_description.__hash__())
    examples = "\n".join([f"Inputs:\nArgs: {align['args']}\nKwargs: {align['kwargs']}\nOutput: {align['output']}" for align in align_examples])

def test_parse_align_datasets():
    # Test that all the examples that are aligned are correctly parsable into the prompt format we have defined
    _parse_examples(summarise_list_generic)
    _parse_examples(summarise_list_typing)
    _parse_examples(summarise_list_pydantic)
    _parse_examples(summarise_list_dict)
    _parse_examples(summarise_list_int)
    _parse_examples(summarise_list_Union)
    _parse_examples(summarise_str)
    _parse_examples(summarise_pydantic)
    _parse_examples(summarise_dict)
    _parse_examples(summarise_int)
    _parse_examples(summarise_Union)
    print("All examples parsed correctly!")




from pydantic import Field
from typing import Annotated
from dotenv import load_dotenv
load_dotenv()
import tanuki

@tanuki.patch
def score_sentiment(input: str) -> Annotated[int, Field(gt=0, lt=10)]:
    """
    Scores the input between 0-10
    """


@tanuki.align
def align_score_sentiment():
    """Register several examples to align your function"""

    assert score_sentiment("I love you") == 10
    assert score_sentiment("I hate you") == 0
    assert score_sentiment("You're okay I guess") == 5


# This is a normal test that can be invoked
def test_score_sentiment():
    """We can test the function as normal using Pytest or Unittest"""
    assert score_sentiment("I like you") == 7



import os

from dotenv import load_dotenv

# Load environment variables
load_dotenv()

VERIFIED_SOURCE_EMAIL = os.getenv('VERIFIED_SOURCE_EMAIL')

from boto3.session import Session

def send_email(subject, body, recipient):
    session = Session()
    ses_client = session.client('ses')
    try:
        response = ses_client.send_email(
            Source=VERIFIED_SOURCE_EMAIL,  # This email must be verified with Amazon SES.
            Destination={
                'ToAddresses': [
                    recipient,
                ],
            },
            Message={
                'Subject': {
                    'Data': subject,
                    'Charset': 'UTF-8'
                },
                'Body': {
                    'Text': {
                        'Data': body,
                        'Charset': 'UTF-8'
                    },
                },
            }
        )
        return response
    except Exception as e:
        print(f"An error occurred: {e}")



import os

from bitarray import bitarray

from tanuki.bloom_filter import BloomFilter
from tanuki.models.function_example import FunctionExample
import random
import string

from tanuki.trackers.abc_buffered_logger import EXPECTED_ITEMS, FALSE_POSITIVE_RATE
from tanuki.trackers.filesystem_buffered_logger import FilesystemBufferedLogger

logger = FilesystemBufferedLogger("test")
bloom_filter_persistence = logger.get_bloom_filter_persistence()

def test_add():
    #bloom_filter = BloomFilter(*optimal_bloom_filter_params(EXPECTED_ITEMS, FALSE_POSITIVE_RATE))
    logger = FilesystemBufferedLogger("test")
    example = FunctionExample((0,), {}, 0 * 2)
    example_data = str(example.__dict__).encode('utf-8') + b'\n'
    nr_of_calls = 10
    nr_of_errors = 0
    for _ in range(nr_of_calls):
        # generate a random string of length 10
        random_string = ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(30))
        before_bit_array = logger.bloom_filter.bit_array.copy()

        logger.log_symbolic_patch(f"test_{random_string}", example)

        after_bit_array = logger.bloom_filter.bit_array
        is_same = before_bit_array == after_bit_array
        looked_up = logger.bloom_filter.lookup(f"test_{random_string}_" + example_data.decode('utf-8'))
        if not looked_up or is_same:
            nr_of_errors += 1
    # check duplicate rate is below 20%
    assert nr_of_errors/nr_of_calls <= 0.2

def test_add_lookup():
    bf1 = BloomFilter(
        bloom_filter_persistence,
        expected_number_of_elements=EXPECTED_ITEMS,
        false_positive_probability=FALSE_POSITIVE_RATE)
    example = FunctionExample((0,), {}, 0 * 2)
    nr_of_calls = 10
    nr_of_errors = 0
    for _ in range(nr_of_calls):
        random_string = ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(30))
        bloom_filter_input = f"{random_string}_{str(example.__dict__)}"
        bf1.add(bloom_filter_input)
        looked_up = bf1.lookup(bloom_filter_input)
        if not looked_up:
            nr_of_errors += 1

    # check duplicate rate is below 20%
    assert nr_of_errors/nr_of_calls <= 0.2


def test_bloom_filter_persistence():
    # Create a Bloom filter and add some values
    bf1 = BloomFilter(
        bloom_filter_persistence,
        expected_number_of_elements=EXPECTED_ITEMS,
        false_positive_probability=FALSE_POSITIVE_RATE)
    for i in range(10):
        bf1.add(str(i))

    # Save the Bloom filter's state
    bf1.save()
    saved_bytes = bf1.bit_array.tobytes()

    count_1 = bf1.bit_array.count(1)
    assert count_1 != 0
    num = 50

    # Manual saving
    with open("test.bin", "wb") as f:
        f.write(saved_bytes)

    # Manual loading
    with open("test.bin", "rb") as f:
        manual_loaded_bytes = f.read()

    assert saved_bytes == manual_loaded_bytes


    # Create a new Bloom filter and load its state
    bf2 = BloomFilter(
        bloom_filter_persistence,
        expected_number_of_elements=EXPECTED_ITEMS,
        false_positive_probability=FALSE_POSITIVE_RATE)
    bf2.load()
    loaded_bytes = bf2.bit_array.tobytes()

    assert bf1.bit_array == bf2.bit_array

    for i, (b1, b2) in enumerate(zip(saved_bytes, loaded_bytes)):
        if b1 != b2:
            print(f"Discrepancy at byte {i}: saved = {b1}, loaded = {b2}")

    # Check if the two Bloom filters are equal
    assert saved_bytes == loaded_bytes
    # Bit wise comparison of the two Bloom filters
    for i in range(len(bf1.bit_array)):
        if not bf1.bit_array[i] == bf2.bit_array[i]:
            print(i)
            print(bf1.bit_array[i-10:i+10])
            print(bf2.bit_array[i-10:i+10])
        assert bf1.bit_array[i] == bf2.bit_array[i]

    assert bf1.bit_array == bf2.bit_array

    # Check if the loaded Bloom filter behaves as expected
    for i in range(10):
        assert bf2.lookup(str(i))

    print("Test passed!")


def test_bit_array_length():
    bf1 = BloomFilter(
        bloom_filter_persistence,
        expected_number_of_elements=EXPECTED_ITEMS,
        false_positive_probability=FALSE_POSITIVE_RATE)

    current_dir = os.path.dirname(os.path.realpath(__file__))
    bf1.bit_array[-1] = True
    bf1.save()

    # Print length of bit_array before loading
    print("Length of bit_array before loading:", len(bf1.bit_array))

    # Print size of the saved file
    file_size = os.path.getsize(os.path.join(current_dir, 'bloom_filter_state.bin'))
    print("Size of saved file (in bytes):", file_size)

    bf2 = BloomFilter(
        bloom_filter_persistence,
        expected_number_of_elements=EXPECTED_ITEMS,
        false_positive_probability=FALSE_POSITIVE_RATE)
    bf2.load()

    ld = bf2.bit_array
    # Print length of bit_array after loading
    print("Length of bit_array after loading:", len(bf2.bit_array))

    assert len(bf1.bit_array) == len(bf2.bit_array), "Mismatch in bit array lengths. {} != {}".format(
        len(bf1.bit_array), len(bf2.bit_array))


def test_file_content_consistency():
    bf1 = BloomFilter(
        bloom_filter_persistence,
        expected_number_of_elements=EXPECTED_ITEMS,
        false_positive_probability=FALSE_POSITIVE_RATE)
    current_dir = os.path.dirname(os.path.realpath(__file__))
    bf1.bit_array[-1] = True
    bf1.save()

    # Read back immediately after saving
    with open(os.path.join(logger.log_directory, 'bloom_filter_state.bin'), 'rb') as f:
        bit_array = bitarray()
        bit_array.frombytes(f.read())
        saved_data = bit_array

    bf2 = BloomFilter(
        bloom_filter_persistence,
        expected_number_of_elements=EXPECTED_ITEMS,
        false_positive_probability=FALSE_POSITIVE_RATE)
    bf2.load()

    assert saved_data == bf2.bit_array, "Saved file content doesn't match loaded content."


def test_simple_test_case():
    bf1 = BloomFilter(
        bloom_filter_persistence,
        expected_number_of_elements=EXPECTED_ITEMS,
        false_positive_probability=FALSE_POSITIVE_RATE)
    items = ['test1', 'test2', 'test3']

    for item in items:
        bf1.add(item)

    current_dir = os.path.dirname(os.path.realpath(__file__))
    bf1.save()

    bf2 = BloomFilter(
        bloom_filter_persistence,
        expected_number_of_elements=EXPECTED_ITEMS,
        false_positive_probability=FALSE_POSITIVE_RATE)
    bf2.load()

    for item in items:
        assert bf2.lookup(item), f"Item {item} not found after loading."


def test_multiple_loggers():
    # test out multiple loggers to ensure bloom filter is saved
    example = FunctionExample((0,), {}, 0 * 2)
    example_data = str(example.__dict__).encode('utf-8') + b'\n'
    nr_of_calls = 10
    nr_of_errors = 0
    for _ in range(nr_of_calls):
        logger_1 = FilesystemBufferedLogger("test")
        # generate a random string of length 10
        random_string = ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(30))
        logger_1.log_symbolic_patch(f"test_{random_string}", example)
        logger_2 = FilesystemBufferedLogger("test")
        looked_up = logger_2.bloom_filter.lookup(f"test_{random_string}_" + example_data.decode('utf-8'))
        if not looked_up:
            nr_of_errors += 1
    # check duplicate rate is below 20%
    assert nr_of_errors/nr_of_calls <= 0.2


if __name__ == "__main__":
    test_bit_array_length()
    test_file_content_consistency()
    test_bloom_filter_persistence()
    test_add_lookup()
    test_add()
    test_multiple_loggers()


from typing import Literal

EXAMPLE_ELEMENT_LIMIT = 1000

# These represent the file extensions for the symbolic patch and alignment datasets
PATCHES = "patches"
PATCH_FILE_EXTENSION_TYPE = Literal[".patches"]
PATCH_FILE_EXTENSION: PATCH_FILE_EXTENSION_TYPE = f".patches"

SYMBOLIC_ALIGNMENTS = "alignments"
ALIGN_FILE_EXTENSION_TYPE = Literal[".alignments"]
ALIGN_FILE_EXTENSION: ALIGN_FILE_EXTENSION_TYPE = f".alignments"

# These represent the file extensions for the embeddable examples positive and negative datasets
POSITIVE_EMBEDDABLE_ALIGNMENTS = "positive"
POSITIVE_FILE_EXTENSION_TYPE = Literal[".positive"]
POSITIVE_FILE_EXTENSION: POSITIVE_FILE_EXTENSION_TYPE = ".positive"

NEGATIVE_EMBEDDABLE_ALIGNMENTS = "negative"
NEGATIVE_FILE_EXTENSION_TYPE = Literal[".negative"]
NEGATIVE_FILE_EXTENSION: NEGATIVE_FILE_EXTENSION_TYPE = ".negative"

# Bloom filter default config
EXPECTED_ITEMS = 10000
FALSE_POSITIVE_RATE = 0.01

# The name of the library
LIB_NAME = "tanuki"
ENVVAR = "TANUKI_LOG_DIR"


import json
from abc import abstractmethod
from typing import Dict, Any, Literal

from tanuki.bloom_filter import BloomFilter
from tanuki.constants import EXPECTED_ITEMS, FALSE_POSITIVE_RATE, ALIGN_FILE_EXTENSION, \
    POSITIVE_FILE_EXTENSION, NEGATIVE_FILE_EXTENSION, PATCH_FILE_EXTENSION
from tanuki.persistence.filter.bloom_interface import IBloomFilterPersistence
from tanuki.trackers.dataset_worker import DatasetWorker

# PATCH_FILE_EXTENSION_TYPE = Literal[".patches"]
# ALIGN_FILE_EXTENSION_TYPE = Literal[".alignments"]
# POSITIVE_EMBEDDING_FILE_EXTENSION_TYPE = Literal[".positive_embedding"]
# NEGATIVE_EMBEDDING_FILE_EXTENSION_TYPE = Literal[".negative_embedding"]
#
# PATCH_FILE_EXTENSION: PATCH_FILE_EXTENSION_TYPE = ".patches"
# ALIGN_FILE_EXTENSION: ALIGN_FILE_EXTENSION_TYPE = ".alignments"
# POSITIVE_EMBEDDING_FILE_EXTENSION: POSITIVE_EMBEDDING_FILE_EXTENSION_TYPE = ".contrastive_positives"
# NEGATIVE_EMBEDDING_FILE_EXTENSION: NEGATIVE_EMBEDDING_FILE_EXTENSION_TYPE = ".contrastive_negatives"
#
# EXPECTED_ITEMS = 10000
# FALSE_POSITIVE_RATE = 0.01
# LIB_NAME = "tanuki"
# ENVVAR = "TANUKI_LOG_DIR"


class ABCBufferedLogger(DatasetWorker):
    def __init__(self, name, level=15):
        self.buffers = {}
        self.mapped_files = {}
        self.miss_count = 0
        self.hit_count = 0
        self.flush_limit = {}
        self.buffer_rolling_size = {}
        self.write_count = 0
        self.write_limit = 1000  # Save the Bloom filter every 1000 writes

        super().__init__(name, level)
        self.bloom_filter = self.create_bloom_filter()
        self.load_bloom_filter()

        self.default_function_config = {"distilled_model": "",
                                        "current_model_stats": {
                                            "trained_on_datapoints": 0,
                                            "running_faults": []},
                                        "last_training_run": {"trained_on_datapoints": 0},
                                        "current_training_run": {},
                                        "teacher_models": ["gpt-4", "gpt-4-32k"],  # currently supported teacher models
                                        "nr_of_training_runs": 0}

    @abstractmethod
    def get_bloom_filter_persistence(self) -> IBloomFilterPersistence:
        """
        Get an instance of the bloom filter persistence provider. This exposes some persistent file storage,
        that must support reading and writing raw byte streams.
        :return:
        """
        pass

    @abstractmethod
    def load_existing_datasets(self) -> Dict[str, Dict[str, Any]]:
        """
        Get the lengths of all datasets backing the registered functions, including aligns.
        :return:
        """
        pass

    @abstractmethod
    def ensure_persistence_location_exists(self):
        """
        Ensure that the place we will be writing to actually exists. If not, create it.
        """
        pass

    @abstractmethod
    def get_patch_location_for_function(self, func_hash, extension="") -> str:
        """
        Get the address of the function patch file.
        :param func_hash: The representation of the function
        :param extension: Whether this is a patch or an alignment
        :return:
        """
        pass

    @abstractmethod
    def write(self, path, data, mode="a") -> None:
        pass

    @abstractmethod
    def read(self, path) -> str:
        pass

    @abstractmethod
    def get_hash_from_path(self, path) -> str:
        pass

    @abstractmethod
    def does_object_exist(self, path) -> bool:
        pass

    def create_bloom_filter(self):
        bloom_filter_persistence = self.get_bloom_filter_persistence()

        bloom_filter = BloomFilter(
            bloom_filter_persistence,
            expected_number_of_elements=EXPECTED_ITEMS,
            false_positive_probability=FALSE_POSITIVE_RATE)
        return bloom_filter

    def load_bloom_filter(self):
        try:
            self.bloom_filter.load()
        except FileNotFoundError:
            self.debug("No Bloom filter found. Creating a new one.")

    def write_symbolic_align_call(self, func_hash, example) -> bool:
        log_file_path = self.get_patch_location_for_function(func_hash, extension=ALIGN_FILE_EXTENSION)
        try:
            # Now, write to the file
            dumpable_object = str(example.__dict__)
            self.write(log_file_path, dumpable_object + "\n", mode="a")
            return True
        except Exception as e:
            return False

    def write_embeddable_align_call(self, func_hash, example, positive=True) -> bool:
        if positive:
            log_file_path = self.get_patch_location_for_function(func_hash, extension=POSITIVE_FILE_EXTENSION)
        else:
            log_file_path = self.get_patch_location_for_function(func_hash, extension=NEGATIVE_FILE_EXTENSION)

        try:
            # Now, write to the file
            dumpable_object = str(example.__dict__)
            self.write(log_file_path, dumpable_object + "\n", mode="a")
            return True
        except Exception as e:
            return False

    def log_embeddable_align(self, func_hash, example, positive=True, **kws):
        """
        Log a contrastive function invocation
        Args:
            func_hash: A string representation of the function signature and input parameters
            example: The example object
            positive: Whether the example is positive or negative
            **kws:
        """
        successfully_saved, new_datapoint = False, False
        try:
            self.ensure_persistence_location_exists()
        except Exception as e:
            return successfully_saved, new_datapoint

        # prepend the function hash to the example
        bloom_filter_representation = func_hash + '_' + str(example.__dict__) + '\n'
        # Check Bloom Filter
        if self.bloom_filter.lookup(bloom_filter_representation):
            return successfully_saved, new_datapoint
        new_datapoint = True
        # add to bloom filter
        self.bloom_filter.add(bloom_filter_representation)
        self.save_bloom_filter()

        successfully_saved = self.write_embeddable_align_call(func_hash, example, positive)
        return successfully_saved, new_datapoint

    def log_symbolic_align(self, func_hash, *args, **kws):
        """
        Log an align function invocation to the file system
        :param func_hash: A string representation of the function signature and input parameters
        :param args: Example objects
        :param kws:
        :return:
        """
        successfully_saved, new_datapoint = False, False
        try:
            self.ensure_persistence_location_exists()
        except Exception as e:
            return successfully_saved, new_datapoint

        example = args[0]

        # prepend the function hash to the example
        bloom_filter_representation = func_hash + '_' + str(example.__dict__) + '\n'
        # Check Bloom Filter
        if self.bloom_filter.lookup(bloom_filter_representation):
            return successfully_saved, new_datapoint
        new_datapoint = True
        # add to bloom filter
        self.bloom_filter.add(bloom_filter_representation)
        self.save_bloom_filter()

        successfully_saved = self.write_symbolic_align_call(func_hash, example)
        return successfully_saved, new_datapoint

    def log_symbolic_patch(self, func_hash, example):
        """
        Log a patched function invocation to the file system
        :param func_hash: A string representation of the function signature and input parameters
        :param example:
        :return:
        """
        if not isinstance(func_hash, str):
            func_hash = str(func_hash)

        example_data = str(example.__dict__).encode('utf-8') + b'\n'

        bloom_filter_representation = func_hash + '_' + example_data.decode('utf-8')
        # Check Bloom Filter
        if self.bloom_filter.lookup(bloom_filter_representation):
            self.hit_count += 1
            return {}

        self.miss_count += 1
        # Add to Bloom Filter
        self.bloom_filter.add(bloom_filter_representation)

        try:
            self.ensure_persistence_location_exists()
        except Exception as e:
            return {}

        log_file_path = self.get_patch_location_for_function(func_hash, extension=PATCH_FILE_EXTENSION)

        if log_file_path not in self.buffers:
            self.buffers[log_file_path] = bytearray()

        if log_file_path not in self.flush_limit:
            self.flush_limit[log_file_path] = 1

        self.buffers[log_file_path].extend(example_data)

        self.write_count += 1
        if log_file_path not in self.buffer_rolling_size:
            self.buffer_rolling_size[log_file_path] = 1
        else:
            self.buffer_rolling_size[log_file_path] += 1
        if self.write_count >= self.write_limit:
            written_datapoints = self.flush()
            self.save_bloom_filter()
            self.write_count = 0  # Reset counter
            return written_datapoints

        if len(self.buffers[log_file_path]) >= min(self.flush_limit[log_file_path], 4096):  # Flush after reaching 4KB
            written_datapoints = {}
            try:
                self.write(log_file_path, self.buffers[log_file_path], mode="a+b")
                # update buffers
                written_datapoints[func_hash] = self.buffer_rolling_size[log_file_path]
                self.buffers[log_file_path].clear()
                self.buffer_rolling_size[log_file_path] = 0
                self.flush_limit[log_file_path] = 2 * self.flush_limit[log_file_path]
                self.save_bloom_filter()
            except Exception as e:
                pass

            return written_datapoints
        return {}

    def save_bloom_filter(self):
        try:
            self.bloom_filter.save()
        except Exception as e:
            self.warning("Could not save Bloom filter: {}".format(e))

    def flush(self):
        # get log directory
        written_datapoints = {}
        for log_file_path, buffer in self.buffers.items():
            if len(buffer) > 0:
                try:
                    self.write(log_file_path, buffer, mode="a+b")
                    written_datapoints[self.get_hash_from_path(log_file_path)] = self.buffer_rolling_size[log_file_path]
                    self.buffer_rolling_size[log_file_path] = 0
                    buffer.clear()
                except Exception as e:
                    pass
        return written_datapoints

    def load_function_config(self, func_hash):

        """
        Get the config file for the function. Uses the message and log directory
        Config file has to be in .json
        """
        default = False
        try:  # try to get the config from the disk. If inaccessible, create a new default one

            self.ensure_persistence_location_exists()
            log_file_path = self.get_patch_location_for_function(func_hash)

            config_path = f"{log_file_path}.json"
            if not self.does_object_exist(config_path):
                function_config = self.default_function_config
                default = True
                self.write_json(config_path, function_config)
            else:
                function_config = self.read_json(config_path)

        except Exception as e:
            function_config = self.default_function_config
            default = True
        return function_config, default

    def update_function_config(self, func_hash, config_to_be_saved):
        """
        Save the config file
        """
        log_file_path = self.get_patch_location_for_function(func_hash)
        config_path = f"{log_file_path}.json"
        try:
            self.write_json(config_path, config_to_be_saved)
        except Exception as e:
            pass

    def write_json(self, path, data):
        self.write(path, json.dumps(data))

    def read_json(self, path):
        return json.loads(self.read(path))



from abc import abstractmethod
from logging import Logger

from tanuki.models.function_example import FunctionExample


class DatasetWorker(Logger):
    def __init__(self, name, level=15):
        super().__init__(name, level)


    @abstractmethod
    def load_existing_datasets(self):
        """
        Get all existing datasets for functions
        Output must be a dictionary with the following structure:
        {
            "alignments": {
                "func_hash": -1
            },
            "patches": {
                "func_hash": -1
            }
        }
        Returns:
            dict: dictionary with the structure above 
        """
        pass

    @abstractmethod
    def log_symbolic_align(self, func_hash, *args, **kws):
        """
        Log an alignment statement to the dataset defined by func_hash
        Args:
            func_hash (str): the function hash
            *args: the args for the datapoint, where args[0] is a FunctionExample(args, kwargs, output) object
            **kws: the kwargs for the datapoint
        """
        pass

    @abstractmethod
    def log_embeddable_align(self, func_hash, first_example: FunctionExample, second_example: FunctionExample, **kws):
        """
        Log a contrastive statement to the dataset defined by func_hash
        Args:
            second_example:
            first_example:
            func_hash (str): the function hash
            *args: the args for the datapoint, where args[0] is a FunctionExample(args, kwargs, output) object
            **kws: the kwargs for the datapoint
        """
        pass

    @abstractmethod
    def log_symbolic_patch(self, func_hash, example):
        """
        Save the example to the patch dataset for the function hash
        Output must be a dictionary with the following structure:
        {
            "func_hash": int
        }
        Where func_hash is the hash of the function and int is the number of datapoints written to the dataset for this function
        
        Args:
            func_hash (str): the function hash
            example (FunctionExample): the example to be saved
        
        Returns:
            dict: dictionary with the structure above
        
        """

    @abstractmethod
    def load_function_config(self, func_hash):

        """
        Get the config file for the function.
        Function config must be a dictionary and have the following structure:
            distilled_model (str): distilled_model_name ("" if no distilled model),
            current_model_stats (dict): dict for current model stats
                example:
                {
                    "trained_on_datapoints" (int): 12 (number of datapoints trained on, 0 if not trained yet),
                    "running_faults" (list): [0, 0, 1] (list of 0s and 1s, where 0 is no fault and 1 is fault)
                }
            
            last_training_run (dict): dict for the last training run
                example:
                {
                    "job_id" (str): job_id for last training run,
                    "trained_on_datapoints" (int): dataset_size that was trained on,
                    "last_checked" (datetime in "%Y-%m-%d %H:%M:%S"): When the last check was made for status of training run)
                }
                Example when no training has been done yet:
                {
                    "trained_on_datapoints": 0
                }

            current_training_run (dict): Same structure as last_training_run, only is non-empty if currently a model is training
                Example when no training has been done yet:
                {}

            teacher_models (list of string): list of teacher models
                example:
                ["gpt-4", "gpt-4-32k"]

            nr_of_training_runs (int): number of training runs that have been done in total
            }
        
        The config file must be returned as a dictionary

        Args:
            func_hash (str): the function hash
        Returns:
            dict: the function config
        """
        pass

    @abstractmethod
    def load_dataset(self, dataset_type, func_hash, return_type):
        """
        Load the datasets for a function hash
        The datasets loaded must be a string, where different datapoints are on new lines
        The output depends on the return type, is it either dataset, dataset_length or both
        
        Args:
            dataset_type (str): either "alignments" or "patches"
            func_hash (str): the function hash
            return_type (str): the return type, either "dataset", "length" or "both"
        Returns:
            tuple: tuple of the form (dataset, length) if return_type is "both", otherwise just dataset or length
        """
        pass

    @abstractmethod
    def update_function_config(self, func_hash, config_to_be_saved):
        """
        Save the config file using the function hash to data storage
        Args:
            func_hash (str): the function hash
            config_to_be_saved (dict): the config to be saved
        
        """
        pass



from main import match_email, wrap_match_email


NAMES = [
    "John Smith",
    "Emily Johnson",
    "Michael Davis",
    "Sarah Wilson",
    "Sarah Miller",
    "David Anderson",
    "Jennifer Brown",
    "Robert Lee",
    "Linda Taylor",
    "William Jones",
    "Mary Jackson",
    "James White",
    "Patricia Martinez",
    "Daniel Wilson",
    "Susan Clark",
    "Joseph Garcia",
    "Karen Harris",
    "Richard Miller",
    "Nancy Thomas",
    "Charles Lewis",
    "Jennifer Allen",
    "Thomas Walker",
    "Margaret Wright",
    "Christopher Scott",
    "Betty Hall",
    "Daniel Moore",
    "Dorothy King",
    "Matthew Turner",
    "Lisa Green",
    "Donald Baker",
    "Helen Adams",
    "Mark Turner",
    "Nancy Collins",
    "Paul Davis",
    "Sharon Edwards",
    "George Taylor",
    "Cynthia Martinez",
    "Kenneth Jackson",
    "Angela Clark",
    "Steven Allen",
    "Donna Miller",
    "Edward Harris",
    "Ruth Wilson",
    "Brian Young",
    "Deborah Martin",
    "Ronald Anderson",
    "Jessica Thompson",
    "Gary Hernandez",
    "Lisa Davis",
    "Jerry Parker",
    "Kimberly Johnson",
    "Sarah Edward",
]


def test_match_email() -> None:
    assert match_email("ava.anderson@cool.com", ["Ava Anderson"]) == ["Ava Anderson"]
    assert match_email("aanderson@cool.com", ["Ava Anderson", "Alan Anderson"]) == ["Ava Anderson", "Alan Anderson"]


def test_match_email_none() -> None:
    assert not match_email("bob@cool.ai", ["Emily Johnson"])


def test_wrap_match_email() -> None:
    assert wrap_match_email("eharris@example.ai", NAMES) == "Edward Harris"
    assert wrap_match_email("sedward@example.com", NAMES) == "Sarah Edward"
    assert wrap_match_email("l.davis@example.com", NAMES) == "Lisa Davis"

    assert not wrap_match_email("davis@example.com", NAMES)
    assert not wrap_match_email("sarah@test.com", NAMES)



from distutils.core import setup

setup(
    name='tanuki.py',  # How you named your package folder (MyLib)
    packages=['tanuki'],  # Chose the same as "name"
    version='0.1.0',  # Start with a small number and increase it with every change you make
    license='MIT',  # Chose a license from here: https://help.github.com/articles/licensing-a-repository
    description='The easiest way to build scalable LLM-powered applications, which gets cheaper and faster over time.',
    # Give a short description about your library
    author='Jack Hopkins',  # Type in your name
    author_email='jack.hopkins@me.com',  # Type in your E-Mail
    url='https://github.com/Tanuki/tanuki.py',  # Provide either the link to your github or to your website
    download_url='https://github.com/Tanuki/tanuki.py/archive/v0.1.0.tar.gz',  # I explain this later on
    keywords=['python', 'ai', 'tdd', 'alignment', 'tanuki', 'distillation', 'pydantic', 'gpt-4', 'llm', 'chat-gpt', 'gpt-4-api', 'ai-functions'],  # Keywords that define your package best
    install_requires=[
        "appdirs~=1.4.4",
        "openai==0.28.1",
        "numpy~=1.26.1",
        "python-dotenv==1.0.0",
        "bitarray==2.8.2",
        "pydantic==2.4.2",
        "requests~=2.31.0"
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Intended Audience :: Developers',  # Define that your audience are developers
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',  # Again, pick a license
        'Programming Language :: Python :: 3',  # Specify which python versions that you want to support
    ],
)



from typing import List

from tanuki.function_modeler import FunctionModeler
from tanuki.register import Register
from tanuki.trackers.filesystem_buffered_logger import FilesystemBufferedLogger
from tanuki.utils import encode_int, decode_int


def dummy_func(input: str) -> List[str]:
    """
    Below you will find an article with stocks analysis. Bring out the stock symbols of companies who are expected to go up or have positive sentiment
    """

def initiate_test(func_modeler, func_hash):
    # initiate the config
    _ = func_modeler.load_function_config(func_hash, )
    for keys, values in func_modeler.function_configs.items():
        if func_hash in keys:
            values["distilled_model"] = "test_ft_1"
            values["teacher_models"] = ["gpt-4","gpt-4-32k"] # model and its token limit]
    func_modeler._update_config_file(func_hash)

def test_encoding():
    ints = []
    characters = []
    for i in range(37):
        character = encode_int(i)
        assert character not in characters
        characters.append(character)
        integer = decode_int(character)
        assert integer not in ints
        ints.append(integer)
        assert i == integer


def test_encode_decode_hash():
    nr_of_training_runs = 5
    workspace_id = 12
    function_description = function_description = Register.load_function_description(dummy_func)
    logger = FilesystemBufferedLogger("test")
    func_modeler = FunctionModeler(logger, environment_id=workspace_id)
    finetune_hash = function_description.__hash__(purpose = "finetune") + encode_int(func_modeler.environment_id) + encode_int(nr_of_training_runs)
    finetune = {"fine_tuned_model": f"Test_model:__{finetune_hash}:asd[]asd",}
    config = func_modeler._construct_config_from_finetune(finetune_hash[:-1], finetune)
    assert config["distilled_model"] == f"Test_model:__{finetune_hash}:asd[]asd"
    assert config["current_model_stats"]["trained_on_datapoints"] == 6400
    assert config["last_training_run"]["trained_on_datapoints"] == 6400
    assert config["teacher_models"] == ["gpt-4","gpt-4-32k"]
    assert config["nr_of_training_runs"] == nr_of_training_runs + 1


if __name__ == '__main__':
    #test_token_counter_finetunable()
    #test_token_counter_non_finetunable_1()
    #test_token_counter_non_finetunable_2()
    #test_error_raise()
    test_encoding()
    test_encode_decode_hash()


import openai
import os
from dotenv import load_dotenv
from typing import List, Optional

load_dotenv()

import tanuki

openai.api_key = os.getenv("OPENAI_API_KEY")


@tanuki.patch
def match_email(email: str, names: List[str]) -> Optional[List[str]]:
    """
    Examine the list of names and return all the names that are likely to correspond to the email address.
    The names array includes both first names and last names. Make sure to consider both the first and last name
    as unique values when searching for a match. Be very conservative in what you consider to be a match.
    If there is one clear match, return a list with one name. If there are multiple clear matches, return a list with all the possible matches.
    If there are no possible matches, return None.
    """


@tanuki.align
def align_match_email() -> None:

    # 1:1 Matching
    assert match_email("john.smith@gmail.com", ["John Smith"]) == ["John Smith"]
    assert match_email("jsmith@google.com", ["John Smith"]) == ["John Smith"]
    assert match_email("j.smith@google.com", ["John Smith"]) == ["John Smith"]
    assert match_email("jasmith@google.com", ["John Smith"]) == ["John Smith"]
    assert match_email("t.swift@yahoo.com", ["John Smith", "Taylor Swift", "Satia Swift", "Milton Swift"]) == ["Taylor Swift"]
    
    # 1: Many Matching
    assert match_email("john.smith@gmail.com", ["John Smith", "Jen Smith"]) == ["John Smith"]

    # Multiple matches
    assert match_email("jsmith@google.com", ["John Smith", "Ella Oakly", "George Bush", "Jaden Smith"]) == ["John Smith", "Jaden Smith"]
    assert match_email("George@google.com", ["John Smith", "George Oakly", "George Bush", "Jaden Smith"]) == ["George Oakly", "George Bush"]

    # Return None because there are no possible matches
    assert match_email("rsunak@gmail.com", ["John Smith", "Jen Smith", "Emily Johnson"]) == None
    assert match_email("marcopolo@gmail.com", ["John Smith", "John Smith Jr.", "David Anderson"]) == None
    assert match_email("johnsmith123@gmail.com", ["Lando Norrs", "David Singh"]) == None


def wrap_match_email(email: str, names: List[str]) -> Optional[str]:
    """
    Wrapper method to call `match_email` method multiple times.
    For this particular use-case, the filter function (match_email) needs to be called atleast twice as GPT4 has quite a high False-Positive rate when working with longer lists of names.
    Calling the filtering function twice worked well to reduce the final False-Positive rate to a acceptable minimum
    """

    match = match_email(email, names) # first call, may include false-positives

    if match:
        match_revised = match_email(email, match) # second call to remove false positives
        if match_revised and len(match_revised) == 1:
            return match_revised[0]

    return None

# Example usage
if __name__ == '__main__':
    print("Aligning...")
    align_match_email()

    print("Matching...")
    match = wrap_match_email("ethan.brown@gmail.com", ["Ethan Brown", "John Smith", "Mary Johnson", "Eva Brown"])
    print(match)

    match = wrap_match_email("ebrown@gmail.com", ["Ethan Brown", "Eva Brown"])
    print(match)



from numpy import square
import openai
import os
from dotenv import load_dotenv
from pydantic import BaseModel
from typing import List, Optional

load_dotenv()

import tanuki
from utils import scrape_url


openai.api_key = os.getenv("OPENAI_API_KEY")


class Property(BaseModel):
    neighborhood: str
    address: str
    price: float
    fee: bool
    beds: float
    bath: float
    listed_by: str


@tanuki.patch
def extract_property(content: str) -> Optional[Property]:
    """
    Examine the content string and extract the rental property details for the neighborhood, address,
    price, number of beds, number of bathrooms, square footage, and company that is listing the property.
    """

@tanuki.align
def align_extract_property() -> None:
    print("Aligning...")
    unit_one = "Rental Unit in Lincoln Square\n      \n\n\n229 West 60th Street #7H\n\n\n\n$7,250\nNO FEE\n\n\n\n\n\n\n\n\n2 Beds\n\n\n\n\n2 Baths\n\n\n\n\n\n                1,386\n                square feet\nft²\n\n\n\n\n\n        Listing by Algin Management"
    assert extract_property(unit_one) == Property(
        neighborhood="Lincoln Square",
        address="229 West 60th Street #7H",
        price=7250.0,
        fee=False,
        beds=2.0,
        bath=2.0,
        listed_by="Algin Management",
    )


if __name__ == '__main__':

    # Align the function
    align_extract_property()

    # Web scrape the url and extract the rental property details
    url = "https://streeteasy.com/2-bedroom-apartments-for-rent/manhattan?page=2"
    contents = scrape_url(url=url, class_name="listingCardBottom")
    print(contents)

    # Process the rental property block using Tanuki
    units = []
    for content in contents[1:3]:
        units.append(extract_property(content))
    print(units)



from tanuki.bloom_filter import BloomFilter
from tanuki.trackers.filesystem_buffered_logger import FilesystemBufferedLogger


def test_hash():
    logger = FilesystemBufferedLogger("test")
    bloom_filter_persistence = logger.get_bloom_filter_persistence()
    bloom = BloomFilter(bloom_filter_persistence, expected_number_of_elements=100, false_positive_probability=0.01)
    assert bloom.hash_functions("test") == bloom.hash_functions("test")

if __name__ == "__main__":
    test_hash()


import os
from typing import List

import openai
from dotenv import load_dotenv
load_dotenv()
import tanuki

@tanuki.patch
def extract_stock_winners_good(input: str) -> List[str]:
    """
    Below you will find an article with stocks analysis. Bring out the stock symbols of companies who are expected to go up or have positive sentiment
    """


@tanuki.align
def test_stock():
    """We can test the function as normal using Pytest or Unittest"""

    input_1 = "Consumer spending makes up a huge fraction of the overall economy. Investors are therefore always looking at consumers to try to gauge whether their financial condition remains healthy. That's a big part of why the stock market saw a bear market in 2022, as some feared that a consumer-led recession would result in much weaker business performance across the sector.\nHowever, that much-anticipated recession hasn't happened yet, and there's still plenty of uncertainty about the future direction of consumer-facing stocks. A pair of earnings reports early Wednesday didn't do much to resolve the debate, as household products giant Procter & Gamble (PG 0.13%) saw its stock rise even as recreational vehicle manufacturer Winnebago Industries (WGO 0.58%) declined."
    assert extract_stock_winners_good(input_1) == ['Procter & Gamble', 'Winnebago Industries']


if __name__ == '__main__':
    test_stock()
    input = "A recent survey by Nationwide, the financial services firm, found that over three-quarters of both Gen Z and millennials expect they will need to continue working into their retirement years because they do not believe Social Security will be enough to rely on in their old age.\nIt's a troubling situation, but the good news is that if you invest in dividend stocks, they can help strengthen your prospects for retirement. Not only can these types of investments increase the value of your portfolio over time, but they will also provide you with recurring cash flow.\nThree dividend stocks that can be excellent investments to include as part of your retirement plan now are UnitedHealth Group (UNH -0.26%), Verizon Communications (VZ 0.83%), and ExxonMobil (XOM 1.31%)."
    output = extract_stock_winners_good(input)
    print(output)


from typing import TypeVar, List, Generic, Union, get_args, get_origin

import numpy as np

T = TypeVar('T')


class Embedding(Generic[T]):
    _data_type = None  # Placeholder for the data type

    def __init__(self, data: List[float]):
        # Determine the origin of the data type (list, np.ndarray, etc.)
        data_type_origin = get_origin(self._data_type) or self._data_type

        if data_type_origin is np.ndarray:
            self._data = np.array(data)
        elif data_type_origin is list:
            # Further check for element type if necessary
            element_type = get_args(self._data_type)[0] if get_args(self._data_type) else None
            if all(isinstance(item, element_type) for item in data):
                self._data = data
            else:
                raise TypeError("Elements of data do not match the expected type")
        elif not data_type_origin:
            self._data = data
        else:
            raise TypeError("Unsupported data type for embedding")

    @classmethod
    def __class_getitem__(cls, item):
        new_cls = type(cls.__name__, (cls,), {'_data_type': item})
        return new_cls

    def __getattr__(self, item):
        return getattr(self._data, item)

    def __repr__(self):
        return repr(self._data)

    def __str__(self):
        return str(self._data)